#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-08-19
#
# Licensed under Apache License, Version 2.0.

import cadquery as cq
import build123d as b3d

import asyncio
import base64
import copy
import os
import pickle
import shutil
import sys
import tempfile

from .render import *
from .plugins import *
from .shape_config import ShapeConfiguration
from .utils import total_size
from . import logging as pc_logging
from . import sync_threads as pc_thread
from . import wrapper

sys.path.append(os.path.join(os.path.dirname(__file__), "wrappers"))
from cq_serialize import register as register_cq_helper


class Shape(ShapeConfiguration):
    name: str
    desc: str
    svg_path: str
    svg_url: str
    # shape: None | b3d.TopoDS_Shape | OCP.TopoDS.TopoDS_Solid

    errors: list[str]

    def __init__(self, config):
        super().__init__(config)
        self.errors = []
        self.lock = asyncio.Lock()
        self.shape = None
        self.components = []
        self.compound = None
        self.with_ports = None

        # Leave the svg path empty to get it created on demand
        self.svg_lock = asyncio.Lock()
        self.svg_path = None
        self.svg_url = None

        self.desc = config.get("desc", None)
        self.desc = self.desc.strip() if self.desc is not None else None

    async def get_components(self):
        if len(self.components) == 0:
            # Maybe it's empty, maybe it's not generated yet
            wrapped = await self.get_wrapped()

            # If it's a compound, we can get the components
            if len(self.components) == 0:
                self.components = [wrapped]

            if self.with_ports is not None:
                ports_list = list(await self.with_ports.get_components())
                if len(ports_list) != 0:
                    self.components.append(ports_list)

        return self.components

    async def get_wrapped(self):
        shape = await self.get_shape()

        # TODO(clairbee): apply 'offset' and 'scale' during instantiation and
        #                 apply to both 'wrapped' and 'components'
        if "offset" in self.config:
            b3d_solid = b3d.Solid.make_box(1, 1, 1)
            b3d_solid.wrapped = shape
            b3d_solid.relocate(b3d.Location(*self.config["offset"]))
            shape = b3d_solid.wrapped
        if "scale" in self.config:
            b3d_solid = b3d.Solid.make_box(1, 1, 1)
            b3d_solid.wrapped = shape
            b3d_solid = b3d_solid.scale(self.config["scale"])
            shape = b3d_solid.wrapped
        return shape

    async def get_cadquery(self) -> cq.Shape:
        cq_solid = cq.Solid.makeBox(1, 1, 1)
        cq_solid.wrapped = await self.get_wrapped()
        return cq_solid

    async def get_build123d(self) -> b3d.Solid:
        b3d_solid = b3d.Solid.make_box(1, 1, 1)
        b3d_solid.wrapped = await self.get_wrapped()
        return b3d_solid

    def regenerate(self):
        """Regenerates the shape generated by AI. Config remains the same."""
        if hasattr(self, "generate"):
            # Invalidate the shape
            # async with self.lock:
            self.shape = None

            # # Truncate the source code file
            # # This will trigger the regeneration of the file on instantiation
            # p = pathlib.Path(self.path)
            # p.unlink(missing_ok=True)
            # p.touch()
            self.generate(self.path)
        else:
            pc_logging.error("No generation function found")

    def do_change(self, change=None):
        if hasattr(self, "change"):
            self.change(self.path, change)
        else:
            pc_logging.error("No change function found")

    async def show_async(self, show_object=None):
        with pc_logging.Action("Show", self.project_name, self.name):
            if show_object is None:
                components = []
                # TODO(clairbee): consider removing this exception handler permanently
                # Comment out the below exception handler for easier troubleshooting in CLI
                try:
                    components = await self.get_components()
                except Exception as e:
                    pc_logging.exception(e)

                if len(components) != 0:
                    import importlib

                    ocp_vscode = importlib.import_module("ocp_vscode")
                    if ocp_vscode is None:
                        pc_logging.warning(
                            'Failed to load "ocp_vscode". Giving up on connection to VS Code.'
                        )
                    else:
                        try:
                            # ocp_vscode.config.status()
                            pc_logging.info(
                                'Visualizing in "OCP CAD Viewer"...'
                            )
                            # pc_logging.debug(self.shape)
                            ocp_vscode.show(
                                *components,
                                progress=None,
                                # TODO(clairbee): make showing (and the connection
                                # to ocp_vscode) a part of the context, and memorize
                                # which part was displayed last. Keep the camera
                                # if the part has not changed.
                                # reset_camera=ocp_vscode.Camera.KEEP,
                            )
                        except Exception as e:
                            pc_logging.warning(e)
                            pc_logging.warning(
                                'No VS Code or "OCP CAD Viewer" extension detected.'
                            )

            if show_object is not None:
                try:
                    shape = await self.get_shape()
                except Exception as e:
                    pc_logging.error(e)
                if shape is not None:
                    show_object(
                        shape,
                        options={},
                    )

    def test(self):
        _ = asyncio.run(self.get_wrapped())

    def show(self, show_object=None):
        asyncio.run(self.show_async(show_object))

    def shape_info(self):
        asyncio.run(self.get_wrapped())
        info = {}
        info["Memory"] = "%.02f KB" % ((total_size(self) + 1023.0) / 1024.0)

        if self.with_ports is not None:
            info["Ports"] = self.with_ports.info()

        return info

    def error(self, msg: str):
        mute = self.config.get("mute", False)
        if mute:
            self.errors.append(msg)
        else:
            pc_logging.error(msg)

    async def render_svg_somewhere(
        self,
        ctx,
        project=None,
        filepath=None,
        line_weight=None,
        viewport_origin=None,
    ):
        """Renders an SVG file somewhere and ignore the project settings"""
        if filepath is None:
            filepath = tempfile.mktemp(".svg")

        obj = await self.get_wrapped()
        if obj is None:
            # pc_logging.error("The shape failed to instantiate")
            self.svg_path = None
            return

        svg_opts, _ = self.render_getopts("svg", ".svg", project, filepath)

        if line_weight is None:
            if "lineWeight" in svg_opts and not svg_opts["lineWeight"] is None:
                line_weight = svg_opts["lineWeight"]
            else:
                line_weight = 1.0

        if viewport_origin is None:
            if (
                "viewportOrigin" in svg_opts
                and not svg_opts["viewportOrigin"] is None
            ):
                viewport_origin = svg_opts["viewportOrigin"]
            else:
                viewport_origin = [100, -100, 100]

        wrapper_path = wrapper.get("render_svg.py")
        request = {
            "wrapped": obj,
            "line_weight": line_weight,
            "viewport_origin": viewport_origin,
        }
        register_cq_helper()
        picklestring = pickle.dumps(request)
        request_serialized = base64.b64encode(picklestring).decode()

        # We don't care about customer preferences much here
        # as this is expected to be hermetic.
        # Stick to the version where CadQuery and build123d are known to work.
        runtime = ctx.get_python_runtime(version="3.10")
        await runtime.ensure("cadquery")  # SVG wrapper requires cq-serialize
        await runtime.ensure("build123d")
        response_serialized, errors = await runtime.run(
            [
                wrapper_path,
                os.path.abspath(filepath),
            ],
            request_serialized,
        )
        sys.stderr.write(errors)

        response = base64.b64decode(response_serialized)
        result = pickle.loads(response)
        if not result["success"]:
            pc_logging.error(
                "RenderSVG failed: %s: %s" % (self.name, result["exception"])
            )
        if "exception" in result and not result["exception"] is None:
            pc_logging.exception(
                "RenderSVG exception: %s" % result["exception"]
            )

        self.svg_path = filepath

    async def _get_svg_path(self, ctx, project):
        async with self.svg_lock:
            if self.svg_path is None:
                await self.render_svg_somewhere(ctx=ctx, project=project)
            return self.svg_path

    def render_getopts(
        self,
        kind,
        extension,
        project=None,
        filepath=None,
    ):
        if not project is None and "render" in project.config_obj:
            render_opts = copy.copy(project.config_obj["render"])
        else:
            render_opts = {}

        if kind in render_opts and not render_opts[kind] is None:
            if isinstance(render_opts[kind], str):
                opts = {"prefix": render_opts[kind]}
            else:
                opts = copy.copy(render_opts[kind])
        else:
            opts = {}

        if (
            "render" in self.config
            and not self.config["render"] is None
            and kind in self.config["render"]
            and not self.config["render"][kind] is None
        ):
            shape_opts = copy.copy(self.config["render"][kind])
            if isinstance(shape_opts, str):
                shape_opts = {"prefix": shape_opts}
            opts = render_cfg_merge(opts, shape_opts)

        # Using the project's config defaults if any
        if filepath is None:
            if "path" in opts and not opts["path"] is None:
                filepath = opts["path"]
            elif "prefix" in opts and not opts["prefix"] is None:
                filepath = opts["prefix"]
            else:
                filepath = "."

            # Check if the format specific section of the config (e.g. "png")
            # provides a relative path and there is output dir in cmd line or
            # the generic section of rendering options in the config.
            if not os.path.isabs(filepath):
                if "output_dir" in render_opts:
                    # TODO(clairbee): consider using project.config_dir
                    # filepath = os.path.join(
                    #     project.config_dir, render_opts["output_dir"], filepath
                    # )
                    filepath = os.path.join(render_opts["output_dir"], filepath)
                elif not project is None:
                    filepath = os.path.join(project.config_dir, filepath)

            if os.path.isdir(filepath):
                filepath = os.path.join(filepath, self.name + extension)

        pc_logging.debug("Rendering: %s" % filepath)

        return opts, filepath

    async def render_svg_async(
        self,
        ctx,
        project=None,
        filepath=None,
    ):
        with pc_logging.Action("RenderSVG", self.project_name, self.name):
            _, filepath = self.render_getopts("svg", ".svg", project, filepath)

            # This creates a temporary file, but it allows to reuse the file
            # with other consumers of self._get_svg_path()
            svg_path = await self._get_svg_path(ctx=ctx, project=project)
            if not svg_path is None and svg_path != filepath:
                if os.path.exists(svg_path):
                    shutil.copyfile(svg_path, filepath)
                else:
                    pc_logging.error("SVG file was not created by the wrapper")

    def render_svg(
        self,
        ctx,
        project=None,
        filepath=None,
    ):
        asyncio.run(self.render_svg_async(ctx, project, filepath))

    async def render_png_async(
        self,
        ctx,
        project=None,
        filepath=None,
        width=None,
        height=None,
    ):
        with pc_logging.Action("RenderPNG", self.project_name, self.name):
            if not plugins.export_png.is_supported():
                pc_logging.error("Export to PNG is not supported")
                return

            png_opts, filepath = self.render_getopts(
                "png", ".png", project, filepath
            )

            if width is None:
                if "width" in png_opts and not png_opts["width"] is None:
                    width = png_opts["width"]
                else:
                    width = DEFAULT_RENDER_WIDTH
            if height is None:
                if "height" in png_opts and not png_opts["height"] is None:
                    height = png_opts["height"]
                else:
                    height = DEFAULT_RENDER_HEIGHT

            # Render the vector image
            svg_path = await self._get_svg_path(ctx=ctx, project=project)

            def do_render_png():
                nonlocal project, svg_path, width, height, filepath
                plugins.export_png.export(
                    project, svg_path, width, height, filepath
                )

            await pc_thread.run(do_render_png)

    def render_png(
        self,
        ctx,
        project=None,
        filepath=None,
        width=None,
        height=None,
    ):
        asyncio.run(
            self.render_png_async(ctx, project, filepath, width, height)
        )

    async def render_step_async(
        self,
        ctx,
        project=None,
        filepath=None,
    ):
        with pc_logging.Action("RenderSTEP", self.project_name, self.name):
            step_opts, filepath = self.render_getopts(
                "step", ".step", project, filepath
            )

            cq_obj = await self.get_cadquery()

            def do_render_step():
                nonlocal project, filepath, cq_obj
                if not project is None:
                    project.ctx.ensure_dirs_for_file(filepath)
                cq.exporters.export(cq_obj, filepath)

            await pc_thread.run(do_render_step)

    def render_step(
        self,
        ctx,
        project=None,
        filepath=None,
    ):
        asyncio.run(self.render_step_async(ctx, project, filepath))

    async def render_stl_async(
        self,
        ctx,
        project=None,
        filepath=None,
        tolerance=None,
        angularTolerance=None,
    ):
        with pc_logging.Action("RenderSTL", self.project_name, self.name):
            stl_opts, filepath = self.render_getopts(
                "stl", ".stl", project, filepath
            )

            if tolerance is None:
                if (
                    "tolerance" in stl_opts
                    and not stl_opts["tolerance"] is None
                ):
                    tolerance = stl_opts["tolerance"]
                else:
                    tolerance = 0.1

            if angularTolerance is None:
                if (
                    "angularTolerance" in stl_opts
                    and not stl_opts["angularTolerance"] is None
                ):
                    angularTolerance = stl_opts["angularTolerance"]
                else:
                    angularTolerance = 0.1

            cq_obj = await self.get_cadquery()

            def do_render_stl():
                nonlocal cq_obj, project, filepath, tolerance, angularTolerance
                if not project is None:
                    project.ctx.ensure_dirs_for_file(filepath)
                cq.exporters.export(
                    cq_obj,
                    filepath,
                    tolerance=tolerance,
                    angularTolerance=angularTolerance,
                )

            await pc_thread.run(do_render_stl)

    def render_stl(
        self,
        ctx,
        project=None,
        filepath=None,
        tolerance=None,
        angularTolerance=None,
    ):
        asyncio.run(
            self.render_stl_async(
                ctx, project, filepath, tolerance, angularTolerance
            )
        )

    async def render_3mf_async(
        self,
        ctx,
        project=None,
        filepath=None,
        tolerance=None,
        angularTolerance=None,
    ):
        with pc_logging.Action("Render3MF", self.project_name, self.name):
            threemf_opts, filepath = self.render_getopts(
                "3mf", ".3mf", project, filepath
            )

            if tolerance is None:
                if (
                    "tolerance" in threemf_opts
                    and not threemf_opts["tolerance"] is None
                ):
                    tolerance = threemf_opts["tolerance"]
                else:
                    tolerance = 0.1

            if angularTolerance is None:
                if (
                    "angularTolerance" in threemf_opts
                    and not threemf_opts["angularTolerance"] is None
                ):
                    angularTolerance = threemf_opts["angularTolerance"]
                else:
                    angularTolerance = 0.1

            cq_obj = await self.get_cadquery()

            def do_render_3mf():
                nonlocal cq_obj, project, filepath, tolerance, angularTolerance
                if not project is None:
                    project.ctx.ensure_dirs_for_file(filepath)
                cq.exporters.export(
                    cq_obj,
                    filepath,
                    tolerance=tolerance,
                    angularTolerance=angularTolerance,
                )

            await pc_thread.run(do_render_3mf)

    def render_3mf(
        self,
        ctx,
        project=None,
        filepath=None,
        tolerance=None,
        angularTolerance=None,
    ):
        asyncio.run(
            self.render_3mf_async(
                ctx, project, filepath, tolerance, angularTolerance
            )
        )

    async def render_threejs_async(
        self,
        ctx,
        project=None,
        filepath=None,
        tolerance=None,
        angularTolerance=None,
    ):
        with pc_logging.Action("RenderThreeJS", self.project_name, self.name):
            threejs_opts, filepath = self.render_getopts(
                "threejs", ".json", project, filepath
            )

            if tolerance is None:
                if (
                    "tolerance" in threejs_opts
                    and not threejs_opts["tolerance"] is None
                ):
                    tolerance = threejs_opts["tolerance"]
                else:
                    tolerance = 0.1

            if angularTolerance is None:
                if (
                    "angularTolerance" in threejs_opts
                    and not threejs_opts["angularTolerance"] is None
                ):
                    angularTolerance = threejs_opts["angularTolerance"]
                else:
                    angularTolerance = 0.1

            cq_obj = await self.get_cadquery()

            def do_render_threejs():
                nonlocal cq_obj, project, filepath, tolerance, angularTolerance
                if not project is None:
                    project.ctx.ensure_dirs_for_file(filepath)
                cq.exporters.export(
                    cq_obj,
                    filepath,
                    tolerance=tolerance,
                    angularTolerance=angularTolerance,
                    exportType=cq.exporters.ExportTypes.TJS,
                )

            await pc_thread.run(do_render_threejs)

    def render_threejs(
        self,
        ctx,
        project=None,
        filepath=None,
        tolerance=None,
        angularTolerance=None,
    ):
        asyncio.run(
            self.render_threejs_async(
                ctx, project, filepath, tolerance, angularTolerance
            )
        )

    async def render_obj_async(
        self,
        ctx,
        project=None,
        filepath=None,
        tolerance=None,
        angularTolerance=None,
    ):
        with pc_logging.Action("RenderOBJ", self.project_name, self.name):
            obj_opts, filepath = self.render_getopts(
                "obj", ".obj", project, filepath
            )

            if tolerance is None:
                if (
                    "tolerance" in obj_opts
                    and not obj_opts["tolerance"] is None
                ):
                    tolerance = obj_opts["tolerance"]
                else:
                    tolerance = 0.1

            if angularTolerance is None:
                if (
                    "angularTolerance" in obj_opts
                    and not obj_opts["angularTolerance"] is None
                ):
                    angularTolerance = obj_opts["angularTolerance"]
                else:
                    angularTolerance = 0.1

            obj = await self.get_wrapped()
            wrapper_path = wrapper.get("render_obj.py")
            request = {
                "wrapped": obj,
                "tolerance": tolerance,
                "angularTolerance": angularTolerance,
            }
            register_cq_helper()
            picklestring = pickle.dumps(request)
            request_serialized = base64.b64encode(picklestring).decode()

            # We don't care about customer preferences much here
            # as this is expected to be hermetic.
            # Stick to the version where CadQuery and build123d are known to work.
            runtime = ctx.get_python_runtime(version="3.10")
            await runtime.ensure("cadquery")
            response_serialized, errors = await runtime.run(
                [
                    wrapper_path,
                    os.path.abspath(filepath),
                ],
                request_serialized,
            )
            sys.stderr.write(errors)

            response = base64.b64decode(response_serialized)
            result = pickle.loads(response)

            if not result["success"]:
                pc_logging.error(
                    "RenderOBJ faled: %s: %s" % (self.name, result["exception"])
                )
            if "exception" in result and not result["exception"] is None:
                pc_logging.exception(result["exception"])

    def render_obj(
        self,
        ctx,
        project=None,
        filepath=None,
        tolerance=None,
        angularTolerance=None,
    ):
        asyncio.run(
            self.render_obj_async(
                ctx, project, filepath, tolerance, angularTolerance
            )
        )

    async def render_gltf_async(
        self,
        ctx,
        project=None,
        filepath=None,
        binary=None,
        tolerance=None,
        angularTolerance=None,
    ):
        with pc_logging.Action("RenderGLTF", self.project_name, self.name):
            gltf_opts, filepath = self.render_getopts(
                "gltf", ".json", project, filepath
            )

            if tolerance is None:
                if (
                    "tolerance" in gltf_opts
                    and not gltf_opts["tolerance"] is None
                ):
                    tolerance = gltf_opts["tolerance"]
                else:
                    tolerance = 0.1

            if angularTolerance is None:
                if (
                    "angularTolerance" in gltf_opts
                    and not gltf_opts["angularTolerance"] is None
                ):
                    angularTolerance = gltf_opts["angularTolerance"]
                else:
                    angularTolerance = 0.1

            if binary is None:
                if "binary" in gltf_opts and not gltf_opts["binary"] is None:
                    binary = gltf_opts["binary"]
                else:
                    binary = False

            b3d_obj = await self.get_build123d()

            def do_render_gltf():
                nonlocal b3d_obj, project, filepath, tolerance, angularTolerance
                b3d.export_gltf(
                    b3d_obj,
                    filepath,
                    binary=binary,
                    linear_deflection=tolerance,
                    angular_deflection=angularTolerance,
                )

            await pc_thread.run(do_render_gltf)

    def render_gltf(
        self,
        ctx,
        project=None,
        filepath=None,
        tolerance=None,
        angularTolerance=None,
    ):
        asyncio.run(
            self.render_gltf_async(
                ctx, project, filepath, tolerance, angularTolerance
            )
        )

    async def render_txt_async(self, ctx, project=None, filepath=None):
        with pc_logging.Action("RenderTXT", self.project_name, self.name):
            if filepath is None:
                filepath = self.path + "/bom.txt"

            if not project is None:
                project.ctx.ensure_dirs_for_file(filepath)
            file = open(filepath, "w+")
            file.write("BoM:\n")
            await self._render_txt_real(file)
            file.close()

    def render_txt(self, ctx, project=None, filepath=None):
        asyncio.run(self.render_txt_async(ctx, project, filepath))

    async def render_markdown_async(self, ctx, project=None, filepath=None):
        with pc_logging.Action("RenderMD", self.project_name, self.name):
            if filepath is None:
                filepath = self.path + "/README.md"

            bom_file = open(filepath, "w+")
            bom_file.write(
                "# "
                + self.name
                + "\n"
                + "## Bill of Materials\n"
                + "| Part | Count* | Vendor | SKU | Preview |\n"
                + "| -- | -- | -- | -- | -- |\n"
            )
            self._render_markdown_real(bom_file)
            bom_file.write(
                """
(\\*) The `Count` field is the number of SKUs to be ordered.
It already takes into account the number of items per SKU.
            """
            )
            bom_file.close()

    def render_markdown(self, ctx, project=None, filepath=None):
        asyncio.run(self.render_markdown_async(ctx, project, filepath))

    async def get_summary_async(self, project=None):
        if "summary" in self.config and not self.config["summary"] is None:
            return self.config["summary"]
        return None

    def get_summary(self, project=None):
        asyncio.run(self.get_summary_async(project))



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-09-07
#
# Licensed under Apache License, Version 2.0.
#

from .provider_data_cart import *
from .provider_request import ProviderRequest


class ProviderRequestQuote(ProviderRequest):
    cart: ProviderCart = None
    result: object = None

    def __init__(self, cart: ProviderCart = ProviderCart()):
        self.cart = cart
        self.result = None

    def set_result(self, result: object):
        self.result = result

    def compose(self):
        composed = {
            "cart": self.cart.compose(),
        }
        if not self.result is None:
            composed["result"] = self.result
        return composed

    def __repr__(self):
        return str(self.compose())



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-08-19
#
# Licensed under Apache License, Version 2.0.

import asyncio
import copy
import typing

import build123d as b3d

from .shape import Shape
from .shape_ai import ShapeWithAi
from . import sync_threads as pc_thread
from . import logging as pc_logging


class AssemblyChild:
    def __init__(self, item, name=None, location=None):
        self.item = item
        self.name = name
        self.location = location


class Assembly(ShapeWithAi):
    path: typing.Optional[str] = None

    def __init__(self, config={}):
        super().__init__(config)

        if "location" in config:
            self.location = config["location"]
        else:
            self.location = None
        self.shape = None
        self.kind = "assemblies"

        # self.children contains all child parts and assemblies before they turn into 'self.shape'
        self.children = []

        # TODO(clairbee): add reference counter to assemblies
        self.count = 0

    async def do_instantiate(self):
        async with self.lock:
            if len(self.children) == 0:
                self.shape = None  # Invalidate if any
                await pc_thread.run(self.instantiate, self)
                if len(self.children) == 0:
                    pc_logging.warning(
                        "The assembly %s:%s is empty"
                        % (self.project_name, self.name)
                    )

    # add is a non-thread-safe method for end users to create custom Assemblies
    def add(
        self,
        child_item: Shape,  # pc.Part or pc.Assembly
        name=None,
        loc=b3d.Location((0.0, 0.0, 0.0), (0.0, 0.0, 1.0), 0.0),
    ):
        self.children.append(AssemblyChild(child_item, name, loc))

        # Keep part reference counter for bill-of-materials purposes
        child_item.ref_inc()

        self.shape = None  # Invalidate if any

    def ref_inc(self):
        for child in self.children:
            child.item.ref_inc()

    async def get_shape(self):
        await self.do_instantiate()
        async with self.lock:
            if self.shape is None:
                child_shapes = []
                tasks = []

                async def per_child(child):
                    item = copy.copy(await child.item.get_build123d())
                    if not child.name is None:
                        item.label = child.name
                    if not child.location is None:
                        item.locate(child.location)
                    return item

                if len(self.children) == 0:
                    pc_logging.warning(
                        "The assembly %s:%s is empty"
                        % (self.project_name, self.name)
                    )
                for child in self.children:
                    tasks.append(per_child(child))

                # TODO(clairbee): revisit whether non-determinism here is acceptable
                for f in asyncio.as_completed(tasks):
                    item = await f
                    child_shapes.append(item)

                compound = b3d.Compound(children=child_shapes)
                if not self.name is None:
                    compound.label = self.name
                if not self.location is None:
                    compound.locate(self.location)
                self.shape = compound.wrapped
            if self.shape is None:
                pc_logging.warning("The shape is None")
            return self.shape
            # return copy.copy(
            #     self.shape
            # )  # TODO(clairbee): fix this for the case when the parts are made with cadquery

    async def get_bom(self):
        await self.do_instantiate()
        async with self.lock:
            bom = {}
            for child in self.children:
                if hasattr(child.item, "get_bom"):
                    # This is an assembly
                    child_bom = await child.item.get_bom()
                    for child_part_name, child_part_count in child_bom.items():
                        if child_part_name in bom:
                            bom[child_part_name] += child_part_count
                        else:
                            bom[child_part_name] = child_part_count
                else:
                    part_name = child.item.project_name + ":" + child.item.name
                    if part_name in bom:
                        bom[part_name] += 1
                    else:
                        bom[part_name] = 1
            return bom

    async def _render_txt_real(self, file):
        await self.do_instantiate()
        for child in self.children:
            child._render_txt_real(file)

    async def _render_markdown_real(self, file):
        await self.do_instantiate()
        for child in self.children:
            child._render_markdown_real(file)



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-03-23
#
# Licensed under Apache License, Version 2.0.
#

import importlib
import re
import threading
import time
from typing import Any

# Lazy-load AI imports as they are not always needed
# import PIL.Image
pil_image = None
# import google.generativeai as google_genai
google_genai = None
# import google.api_core.exceptions
google_api_core_exceptions = None

from . import logging as pc_logging
from .user_config import user_config

lock = threading.Lock()
GOOGLE_API_KEY = None

model_tokens = {
    "gemini-pro": 32600,
    "gemini-pro-vision": 16300,
}


def google_once():
    global GOOGLE_API_KEY
    global pil_image
    global google_genai
    global google_api_core_exceptions

    with lock:
        if pil_image is None:
            pil_image = importlib.import_module("PIL.Image")

        if google_genai is None:
            google_genai = importlib.import_module("google.generativeai")

        if google_api_core_exceptions is None:
            google_api_core_exceptions = importlib.import_module(
                "google.api_core.exceptions"
            )

        latest_key = user_config.google_api_key
        if latest_key != GOOGLE_API_KEY:
            GOOGLE_API_KEY = latest_key
            if not GOOGLE_API_KEY is None:
                google_genai.configure(api_key=GOOGLE_API_KEY)
                return True

        if GOOGLE_API_KEY is None:
            raise Exception("Google API key is not set")

    return True


class AiGoogle:
    def generate_google(
        self,
        model: str,
        prompt: str,
        config: dict[str, Any] = {},
        options_num: int = 1,
    ):
        if not google_once():
            return None

        if "tokens" in config:
            tokens = config["tokens"]
        else:
            tokens = model_tokens[model] if model in model_tokens else None

        if "top_p" in config:
            top_p = config["top_p"]
        else:
            top_p = None

        if "top_k" in config:
            top_k = config["top_k"]
        else:
            top_k = None

        if "temperature" in config:
            temperature = config["temperature"]
        else:
            temperature = None

        image_content = []

        def insert_image(match):
            filename = match.group(1)
            image_content.append(pil_image.open(filename))
            return "IMAGE_INSERTED_HERE"

        prompt = re.sub(r"INSERT_IMAGE_HERE\(([^)]*)\)", insert_image, prompt)
        text_content = prompt.split("IMAGE_INSERTED_HERE")

        content = []
        for i in range(len(text_content)):
            content.append(text_content[i])
            if i < len(image_content):
                content.append(image_content[i])

        client = google_genai.GenerativeModel(
            model,
            generation_config={
                "candidate_count": 1,
                # "candidate_count": options_num,  # TODO(clairbee): not supported yet? not any more?
                "max_output_tokens": tokens,
                "temperature": temperature,
                "top_p": top_p,
                "top_k": top_k,
            },
        )
        candidates = []
        options_left = options_num
        while options_left > 0:
            retry = True
            while retry == True:
                retry = False
                try:
                    response = client.generate_content(content)
                except google_api_core_exceptions.ResourceExhausted as e:
                    pc_logging.exception(e)
                    retry = True
                    time.sleep(60)
                except google_api_core_exceptions.InternalServerError as e:
                    pc_logging.exception(e)
                    retry = True
                    time.sleep(1)

            if response is None:
                error = "%s: Failed to generate content" % self.name
                pc_logging.error(error)
                continue

            options_created = len(response.candidates)
            candidates.extend(response.candidates)
            options_left -= options_created if options_created > 0 else 1

        products = []
        try:
            for candidate in candidates:
                product = ""
                for part in candidate.content.parts:
                    if "text" in part:
                        product += part.text + "\n"

                products.append(product)
        except Exception as e:
            pc_logging.exception(e)

        return products



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-12-30
#
# Licensed under Apache License, Version 2.0.
#

import os


def get(filename):
    return os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "wrappers",
        "wrapper_" + filename,
    )



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-01-06
#
# Licensed under Apache License, Version 2.0.
#

import asyncio
from jinja2 import Environment, FileSystemLoader
import fnmatch
import os
import yaml

from OCP.gp import gp_Trsf
import build123d as b3d

from .assembly import Assembly, AssemblyChild
from .assembly_factory_file import AssemblyFactoryFile
from . import logging as pc_logging
from .utils import normalize_resource_path


class AssemblyFactoryAssy(AssemblyFactoryFile):
    def __init__(self, ctx, source_project, target_project, config):
        with pc_logging.Action("InitASSY", source_project.name, config["name"]):
            super().__init__(
                ctx, source_project, target_project, config, extension=".assy"
            )
            # Complement the config object here if necessary
            self._create(config)

    def instantiate(self, assembly):
        # # This method is best executed on a thread but the current Python version
        # # might not be good enough to do that.
        # try:
        #     # Try running on the current event loop if any.
        #     loop = asyncio.get_running_loop()
        #     # task = loop.create_task(self.instantiate_async(assembly))
        #     # task.
        #     f = asyncio.run_coroutine_threadsafe(self.instantiate_async(assembly), loop)
        #     raise Exception("IT WORKS")
        # except RuntimeError as e:
        #     print(e)
        #     # Running on a dedicated thread and there is no event loop here yet.
        asyncio.run(self.instantiate_async(assembly))

    async def instantiate_async(self, assembly):
        await super().instantiate(assembly)

        with pc_logging.Action("ASSY", assembly.project_name, assembly.name):
            assy = {}
            if os.path.exists(self.path):
                params = {}
                if "parameters" in self.config:
                    for param_name, param in self.config["parameters"].items():
                        params["param_" + param_name] = param["default"]
                params["name"] = self.config["name"]

                # Read the body of the configuration file
                fp = open(self.path, "r")
                config = fp.read()
                fp.close()

                # Resolve Jinja templates
                template = Environment(
                    loader=FileSystemLoader(
                        os.path.dirname(self.path) + os.path.sep
                    )
                ).from_string(config)
                config = template.render(params)

                # Parse the resulting config
                try:
                    assy = yaml.safe_load(config)
                except Exception as e:
                    pc_logging.error(
                        "ERROR: Failed to parse the assembly file %s"
                        % self.path
                    )
                if assy is None:
                    assy = {}
            else:
                pc_logging.error(
                    "ERROR: Assembly file not found: %s" % self.path
                )

            result = await self.handle_node(assembly, assy)
            if not result is None:
                assembly.children.append(
                    AssemblyChild(result[0], result[1], result[2])
                )
                # Keep part reference counter for bill-of-materials purposes
                result[0].ref_inc()
            else:
                pc_logging.warning("Assembly is empty")

            self.ctx.stats_assemblies_instantiated += 1

    async def handle_node_list(self, assembly, node_list):
        tasks = []

        async def wait_for_tasks():
            nonlocal tasks
            while len(tasks) > 0:
                task = tasks.pop(0)
                f = await asyncio.tasks.wait([task])
                result = f[0].pop().result()
                if not result is None:
                    assembly.children.append(
                        AssemblyChild(result[0], result[1], result[2])
                    )
                    # Keep part reference counter for bill-of-materials purposes
                    result[0].ref_inc()

        for link in node_list:
            if "connect" in link or "connectPorts" in link:
                # wait for all previous nodes to get added first
                await wait_for_tasks()
            tasks.append(asyncio.create_task(self.handle_node(assembly, link)))
        await wait_for_tasks()

    async def handle_node(self, assembly, node):
        # "name" is an optional parameter for both parts and assemblies
        if "name" in node:
            name = node["name"]
        else:
            name = None

        connect = None
        connect_with_iface = None
        connect_with_params = None
        connect_with_instance = None
        connect_with_instance_pattern = None
        connect_with_port = None
        connect_with_port_pattern = None
        connect_to_name = None
        connect_to_iface = None
        connect_to_params = None
        connect_to_instance = None
        connect_to_instance_pattern = None
        connect_to_port = None
        connect_to_port_pattern = None
        # "location" is an optional parameter for both parts and assemblies
        if "location" in node:
            l = node["location"]
            location = b3d.Location(
                (l[0][0], l[0][1], l[0][2]), (l[1][0], l[1][1], l[1][2]), l[2]
            )
        elif "connect" in node:
            connect = node["connect"]
            connect_with_iface = connect.get("with", None)
            if connect_with_iface is not None and ":" not in connect_with_iface:
                connect_with_iface = (
                    self.project.name + ":" + connect_with_iface
                )
            connect_with_params = connect.get("withParams", None)
            connect_with_instance = connect.get("withInstance", None)
            connect_with_port = connect.get("withPort", None)

            connect_to_name = connect.get("name", None)

            connect_to_iface = connect.get("to", None)
            if connect_to_iface is not None and ":" not in connect_to_iface:
                connect_to_iface = self.project.name + ":" + connect_to_iface
            connect_to_params = connect.get("toParams", None)
            connect_to_instance = connect.get("toInstance", None)
            connect_to_port = connect.get("toPort", None)
            location = None
        elif "connectPorts" in node:
            connect = node["connectPorts"]
            connect_with_port = connect.get("with", None)
            connect_to_name = connect.get("name", None)
            connect_to_port = connect.get("to", None)
            location = None
        else:
            location = b3d.Location((0, 0, 0), (0, 0, 1), 0)

        if connect_with_instance is not None and "*" in connect_with_instance:
            connect_with_instance_pattern = connect_with_instance
            connect_with_instance = None
        if connect_to_instance is not None and "*" in connect_to_instance:
            connect_to_instance_pattern = connect_to_instance
            connect_to_instance = None

        if connect_with_port is not None and "*" in connect_with_port:
            connect_with_port_pattern = connect_with_port
            connect_with_port = None
        if connect_to_port is not None and "*" in connect_to_port:
            connect_to_port_pattern = connect_to_port
            connect_to_port = None

        # Check if this node is for an assembly
        if "links" in node and not node["links"] is None:
            item = Assembly(
                {"name": name}
            )  # TODO(clairbee): revisit why node["links"]) was used there
            item.instantiate = lambda x: True
            await self.handle_node_list(item, node["links"])
        else:
            # This is a node for a part or an assembly
            params = {}
            if "params" in node:
                for paramName in node["params"]:
                    params[paramName] = node["params"][paramName]

            if "assembly" in node:
                assy_name = node["assembly"]
                if name is None:
                    name = assy_name
                if "package" in node:
                    assy_name = node["package"] + ":" + assy_name
                elif not ":" in assy_name:
                    assy_name = ":" + assy_name
                assy_name = normalize_resource_path(
                    self.project.name, assy_name
                )
                item = self.ctx._get_assembly(assy_name, params)
                if item is None:
                    pc_logging.error("Assembly not found: %s" % name)
                    raise Exception("Assembly not found")
            elif "part" in node:
                part_name = node["part"]
                if name is None:
                    name = part_name
                if "package" in node:
                    part_name = node["package"] + ":" + part_name
                elif not ":" in part_name:
                    part_name = ":" + part_name
                part_name = normalize_resource_path(
                    self.project.name, part_name
                )
                item = self.ctx._get_part(part_name, params)
                if item is None:
                    pc_logging.error(
                        "Part not found: %s in %s" % (name, self.name)
                    )
                    raise Exception(
                        "Part not found: %s in %s" % (name, self.name)
                    )
            else:
                item = None

            if connect is not None:
                pc_logging.debug("Attempting to connect %s" % name)
                source_port = None
                source_iface = None
                source_iface_obj = None
                source_offsets = []
                source_iface_instance = None
                target_part = None
                target_part_location = None
                target_port = None
                target_iface = None
                target_iface_obj = None
                target_offsets = []
                target_iface_instance = None
                # trsf = None # TODO(clairbee): implement offsets

                for child in assembly.children:
                    # if hasattr(child, "name"):
                    #     pc_logging.debug("Found part: %s" % child.name)
                    if hasattr(child, "name") and child.name == connect_to_name:
                        # pc_logging.debug(
                        #     "Found target part: %s" % connect_to_name
                        # )

                        target_part = child.item
                        break

                if target_part is None:
                    pc_logging.error(
                        "Target part not found: %s" % connect_to_name
                    )
                else:
                    if hasattr(child, "location"):
                        target_part_location = child.location
                    else:
                        target_part_location = b3d.Location(
                            (0, 0, 0), (0, 0, 1), 0
                        )

                    # If there is no source interface specified,
                    # but there is only one present, then use it
                    if (
                        connect_with_iface is None
                        and item.with_ports is not None
                        and not "ports" in item.with_ports.config
                        and len(list(item.with_ports.get_interfaces().keys()))
                        == 1
                    ):
                        connect_with_iface = list(
                            item.with_ports.get_interfaces().keys()
                        )[0]
                        pc_logging.debug(
                            "Using the only source interface: %s"
                            % connect_with_iface
                        )

                    # If there is only one source port, then just use it.
                    if connect_with_port is None:
                        if len(list(item.with_ports.get_ports().keys())) == 1:
                            connect_with_port = list(
                                item.with_ports.get_ports().keys()
                            )[0]
                            pc_logging.debug(
                                "Using the only source port: %s"
                                % connect_with_port
                            )
                            # If the instance is known and the port pattern is configured
                        elif connect_with_port_pattern is not None:
                            matched = []
                            for port in item.with_ports.get_ports().values():
                                if fnmatch.fnmatch(
                                    port, connect_with_port_pattern
                                ):
                                    matched.append(port)

                            if len(matched) == 1:
                                connect_with_port = matched[0]
                                pc_logging.debug(
                                    "Found source port by pattern: %s"
                                    % connect_with_port
                                )
                            elif len(matched) > 1:
                                pc_logging.debug(
                                    "Multiple source ports are matching the pattern, pending interface checks: %s"
                                    % matched
                                )

                    # If the source port is known, then use it
                    if not connect_with_port is None:
                        source_port = item.with_ports.get_ports()[
                            connect_with_port
                        ]
                        pc_logging.debug(
                            "Configured source port: %s" % source_port.name
                        )
                    else:
                        # Source port is not configured.
                        # It needs to be determined below.
                        pass

                    # If there is no target interface specified,
                    # but there is only one present, then use it
                    if (
                        connect_to_iface is None
                        and target_part.with_ports is not None
                        and not "ports" in target_part.with_ports.config
                        and len(
                            list(target_part.with_ports.get_interfaces().keys())
                        )
                        == 1
                    ):
                        connect_to_iface = list(
                            target_part.with_ports.get_interfaces().keys()
                        )[0]
                        pc_logging.debug(
                            "Using the only target interface: %s"
                            % connect_to_iface
                        )

                    # If there is only one target port, then just use it.
                    if connect_to_port is None:
                        if (
                            len(list(target_part.with_ports.get_ports().keys()))
                            == 1
                        ):
                            connect_to_port = list(
                                target_part.with_ports.get_ports().keys()
                            )[0]
                            pc_logging.debug(
                                "Using the only target port: %s"
                                % connect_to_port
                            )
                        elif connect_to_port_pattern is not None:
                            matched = []
                            for (
                                port
                            ) in target_part.with_ports.get_ports().keys():
                                if fnmatch.fnmatch(
                                    port, connect_to_port_pattern
                                ):
                                    matched.append(port)

                            if len(matched) == 1:
                                connect_to_port = matched[0]
                                pc_logging.debug(
                                    "Found target port by pattern: %s"
                                    % connect_to_port
                                )
                            elif len(matched) > 1:
                                pc_logging.debug(
                                    "Multiple target ports are matching the pattern, pending interface checks: %s"
                                    % matched
                                )

                    # If the target port is configured, then use it
                    if not connect_to_port is None:
                        target_port = target_part.with_ports.get_ports()[
                            connect_to_port
                        ]
                        pc_logging.debug(
                            "Configured target port: %s" % target_port.name
                        )
                    else:
                        # Target port is not configured.
                        # It needs to be determined below.
                        pass

                    # If we have either source or target interface/port missing,
                    # then we need to use mating information to determine them.
                    if (
                        connect_with_port is None and connect_with_iface is None
                    ) or (connect_to_port is None and connect_to_iface is None):
                        # FIXME(clairbee): Brake the following step up into two:
                        # 1. If we know the port name but not the interface,
                        #    learn which interface it belons to.
                        # 2. Get the list of interface for each part for which
                        #    we don't know the interface yet.
                        # 3. Given two lists of interfaces, find the mating
                        #    candidates.
                        (
                            source_candidate_interfaces,
                            target_candidate_interfaces,
                        ) = self.project.ctx.find_mating_interfaces(
                            item, target_part
                        )

                        # If it's the source interface that we need to auto-detect
                        if (
                            connect_with_port is None
                            and connect_with_iface is None
                        ):
                            if len(source_candidate_interfaces) == 1:
                                connect_with_iface = (
                                    source_candidate_interfaces.pop()
                                )
                            elif len(source_candidate_interfaces) > 1:
                                # FIXME(clairbee): what if we have patterns configured?
                                pc_logging.debug(
                                    "Multiple source interfaces are candidates for connection: %s"
                                    % source_candidate_interfaces
                                )
                            else:
                                pc_logging.debug(
                                    "No source interfaces are candidates for connection"
                                )

                        # If it's the target interface that we need to auto-detect
                        if connect_to_port is None and connect_to_iface is None:
                            if len(target_candidate_interfaces) == 1:
                                connect_to_iface = (
                                    target_candidate_interfaces.pop()
                                )
                            elif len(target_candidate_interfaces) > 1:
                                # FIXME(clairbee): what if we have patterns configured?
                                pc_logging.debug(
                                    "Multiple target interfaces are candidates for connection: %s"
                                    % target_candidate_interfaces
                                )
                            else:
                                pc_logging.debug(
                                    "No target interfaces are candidates for connection"
                                )

                    # Resolve interface names
                    if connect_with_iface is not None:
                        source_iface = item.with_ports.get_interface(
                            connect_with_iface
                        )
                        source_iface_obj = self.project.ctx.get_interface(
                            connect_with_iface
                        )
                    if connect_to_iface is not None:
                        target_iface = target_part.with_ports.get_interface(
                            connect_to_iface
                        )
                        target_iface_obj = self.project.ctx.get_interface(
                            connect_to_iface
                        )

                    # If we know the source interface but not the port
                    if (
                        connect_with_port is None
                        and source_iface_obj is not None
                    ):
                        # If the instance is specified, then find it
                        pc_logging.debug(
                            "Source interface instances: %s"
                            % source_iface.keys()
                        )

                        # If there is an instance pattern configured
                        if (
                            source_iface is not None
                            and source_iface_instance is None
                            and connect_with_instance_pattern is not None
                        ):
                            matched = []
                            for instance in source_iface.values():
                                if fnmatch.fnmatch(
                                    instance, connect_with_instance_pattern
                                ):
                                    matched.append(instance)

                            if len(matched) == 1:
                                connect_with_instance = matched[0]
                                pc_logging.debug(
                                    "Found source instance by pattern: %s"
                                    % connect_with_instance
                                )
                            elif len(matched) > 1:
                                pc_logging.debug(
                                    "Multiple source instances are matching the pattern: %s"
                                    % matched
                                )

                            if connect_with_instance is None:
                                pc_logging.error(
                                    "Connect %s to %s: instance is not found by pattern: %s"
                                    % (
                                        name,
                                        connect_to_name,
                                        connect_with_instance_pattern,
                                    )
                                )

                        if not connect_with_instance is None:
                            if not connect_with_instance in source_iface:
                                pc_logging.error(
                                    "Connect %s to %s: source instance is not found: %s"
                                    % (
                                        name,
                                        connect_to_name,
                                        connect_with_instance,
                                    )
                                )
                            else:
                                source_iface_instance = source_iface[
                                    connect_with_instance
                                ]
                        # If there is only one instance, then use it
                        elif len(list(source_iface.values())) == 1:
                            source_iface_instance = list(source_iface.values())[
                                0
                            ]
                        elif len(list(source_iface.values())) > 1:
                            # This could be ok if we have a port name or pattern
                            pc_logging.debug(
                                "Missing instance specification for the source interface: %s"
                                % connect_with_iface
                            )

                        # If the instance is known and it has only one port, then use it
                        if (
                            source_iface_instance is not None
                            and connect_with_port is None
                            and len(list(source_iface_instance.values())) == 1
                        ):
                            connect_with_port = list(
                                source_iface_instance.values()
                            )[0]
                        # If the instance is known and the port pattern is configured
                        elif (
                            source_iface_instance is not None
                            and connect_with_port is None
                            and connect_with_port_pattern is not None
                        ):
                            matched = []
                            for port in source_iface_instance.values():
                                if fnmatch.fnmatch(
                                    port, connect_with_port_pattern
                                ):
                                    matched.append(port)

                            if len(matched) == 1:
                                connect_with_port = matched[0]
                                pc_logging.debug(
                                    "Found source port by pattern: %s"
                                    % connect_with_port
                                )
                            elif len(matched) > 1:
                                pc_logging.debug(
                                    "Multiple source ports are matching the pattern: %s"
                                    % matched
                                )

                            if connect_with_port is None:
                                pc_logging.error(
                                    "Connect %s to %s: port is not found by pattern: %s"
                                    % (
                                        name,
                                        connect_to_name,
                                        connect_with_port_pattern,
                                    )
                                )
                        # If the instance is known and the interface has the lead port
                        elif (
                            source_iface_instance is not None
                            and connect_with_port is None
                            and source_iface_obj.lead_port is not None
                        ):
                            for port in source_iface_instance.values():
                                if fnmatch.fnmatch(
                                    port, "*" + source_iface_obj.lead_port
                                ):
                                    connect_with_port = port
                                    break
                            if connect_with_port is None:
                                pc_logging.error(
                                    "Connect %s to %s: lead port is not found: %s"
                                    % (
                                        name,
                                        connect_to_name,
                                        source_iface_obj.lead_port,
                                    )
                                )
                        # pc_logging.info("Found source interface: %s" % source_iface.name)

                    # If we know the target interface but not the port
                    if connect_to_port is None and target_iface_obj is not None:
                        # If the instance is specified, then find it
                        pc_logging.debug(
                            "Target interface instances: %s"
                            % target_iface.keys()
                        )

                        # If there is an instance pattern configured and there is only one match
                        if (
                            target_iface is not None
                            and target_iface_instance is None
                            and connect_to_instance_pattern is not None
                        ):
                            matched = []
                            for instance in list(target_iface.keys()):
                                pc_logging.debug("Instance: %s" % instance)
                                pc_logging.debug(
                                    "Pattern: %s" % connect_to_instance_pattern
                                )
                                if fnmatch.fnmatch(
                                    instance, connect_to_instance_pattern
                                ):
                                    matched.append(instance)

                            if len(matched) == 1:
                                connect_to_instance = matched[0]
                                pc_logging.debug(
                                    "Found target instance by pattern: %s"
                                    % connect_to_instance
                                )
                            elif len(matched) > 1:
                                pc_logging.debug(
                                    "Multiple target instances are matching the pattern: %s"
                                    % matched
                                )

                            if connect_to_instance is None:
                                pc_logging.error(
                                    "Connect %s to %s: instance is not found by pattern: %s"
                                    % (
                                        name,
                                        connect_to_name,
                                        connect_to_instance_pattern,
                                    )
                                )

                        if not connect_to_instance is None:
                            if not connect_to_instance in target_iface:
                                pc_logging.error(
                                    "Connect %s to %s: target instance is not found: %s"
                                    % (
                                        name,
                                        connect_to_name,
                                        connect_to_instance,
                                    )
                                )
                            else:
                                target_iface_instance = target_iface[
                                    connect_to_instance
                                ]
                        # If there is only one instance, then use it
                        elif len(list(target_iface.values())) == 1:
                            target_iface_instance = list(target_iface.values())[
                                0
                            ]
                        elif len(list(target_iface.values())) > 1:
                            # This could be ok if we have a port name or pattern
                            pc_logging.debug(
                                "Missing instance specification for the target interface: %s"
                                % connect_to_iface
                            )

                        # If the instance is known and it has only one port, then use it
                        if (
                            target_iface_instance is not None
                            and connect_to_port is None
                            and len(list(target_iface_instance.values())) == 1
                        ):
                            connect_to_port = list(
                                target_iface_instance.values()
                            )[0]
                        # If the instance is known and the port pattern is configured
                        elif (
                            target_iface_instance is not None
                            and connect_to_port is None
                            and connect_to_port_pattern is not None
                        ):
                            matched = []
                            for port in target_iface_instance.values():
                                if fnmatch.fnmatch(
                                    port, connect_to_port_pattern
                                ):
                                    matched.append(port)

                            if len(matched) == 1:
                                connect_to_port = matched[0]
                                pc_logging.debug(
                                    "Found target port by pattern: %s"
                                    % connect_to_port
                                )
                            elif len(matched) > 1:
                                pc_logging.debug(
                                    "Multiple target ports are matching the pattern: %s"
                                    % matched
                                )

                            if connect_to_port is None:
                                pc_logging.error(
                                    "Connect %s to %s: port is not found by pattern: %s"
                                    % (
                                        name,
                                        connect_to_name,
                                        connect_to_port_pattern,
                                    )
                                )
                        # If the instance is known and the interface has the lead port
                        elif (
                            target_iface_instance is not None
                            and connect_to_port is None
                            and target_iface_obj.lead_port is not None
                        ):
                            for port in target_iface_instance.values():
                                if fnmatch.fnmatch(
                                    port, "*" + target_iface_obj.lead_port
                                ):
                                    connect_to_port = port
                                    break
                            if connect_to_port is None:
                                pc_logging.error(
                                    "Connect %s to %s: lead port is not found: %s"
                                    % (
                                        name,
                                        connect_to_name,
                                        target_iface_obj.lead_port,
                                    )
                                )
                        # pc_logging.info("Found target interface: %s" % target_iface.name)

                    # We know both interface instances but we don't know one of the ports
                    # FIXME(claibee): it is currently working only for cases where we don't know both ports
                    # TODO(clairbee): should the following condition drop the check for iface_instance?
                    #                 shouldn't we try to match ports even if there is no mating information?
                    if (
                        source_iface_instance is not None
                        and target_iface_instance is not None
                        # FIXME(clairbee): what if we only know one?
                        and connect_with_port is None
                        and connect_to_port is None
                    ):
                        pc_logging.debug("Trying to match ports by name")
                        # Both interfaces are known but they have more than one port each.
                        # Let's find a matching pair of ports.
                        source_ports = sorted(
                            list(source_iface_instance.keys())
                        )
                        target_ports = sorted(
                            list(target_iface_instance.keys())
                        )
                        if source_ports == target_ports:
                            # FIXME(clairbee): so what? what if we have patterns configured?
                            # The interfaces have the same number of ports and the same names.
                            # We can connect them using any pair of matching ports.
                            connect_with_port = source_iface_instance[
                                source_ports[0]
                            ]
                            connect_to_port = target_iface_instance[
                                target_ports[0]
                            ]
                        else:
                            connect_with_port_index = -1
                            connect_to_port_index = -1
                            if len(source_ports) != 1:
                                if (
                                    connect_with_port_pattern is None
                                    and connect_to_port_pattern is None
                                ):
                                    if len(source_ports) == len(target_ports):
                                        pc_logging.debug(
                                            "Connect %s to %s: port mating is not detected deterministically on BOTH ends, guessing alphabetically..."
                                            % (name, connect_to_name)
                                        )
                                        connect_with_port_index = 0
                                        connect_to_port_index = 0
                                    else:
                                        pc_logging.error(
                                            "Connect %s to %s: port mating is not detected deterministically on BOTH ends, different number of ports"
                                            % (name, connect_to_name)
                                        )

                                if connect_with_port_pattern is not None:
                                    for i, port in enumerate(source_ports):
                                        if fnmatch.fnmatch(
                                            port, connect_with_port_pattern
                                        ):
                                            connect_with_port_index = i
                                            break
                                if connect_to_port_pattern is not None:
                                    for i, port in enumerate(target_ports):
                                        if fnmatch.fnmatch(
                                            port, connect_to_port_pattern
                                        ):
                                            connect_to_port_index = i
                                            break

                                if (
                                    len(source_ports) == len(target_ports)
                                    and connect_with_port_index != -1
                                    and connect_to_port_index == -1
                                ):
                                    pc_logging.debug(
                                        "Connect %s to %s: port selection is not detected on the target end, guessing alphabetically..."
                                        % (name, connect_to_name)
                                    )
                                    connect_to_port_index = (
                                        connect_with_port_index
                                    )
                                if (
                                    len(source_ports) == len(target_ports)
                                    and connect_with_port_index == -1
                                    and connect_to_port_index != -1
                                ):
                                    pc_logging.debug(
                                        "Connect %s to %s: port selection is not detected on the source end, guessing alphabetically..."
                                        % (name, connect_to_name)
                                    )
                                    connect_with_port_index = (
                                        connect_to_port_index
                                    )

                            if connect_with_port_index != -1:
                                connect_with_port = source_iface_instance[
                                    source_ports[connect_with_port_index]
                                ]
                            if connect_to_port_index != -1:
                                connect_to_port = target_iface_instance[
                                    target_ports[connect_to_port_index]
                                ]

                    # If the source port is determined, then use it
                    if not connect_with_port is None and source_port is None:
                        source_port = item.with_ports.get_ports()[
                            connect_with_port
                        ]
                        pc_logging.debug(
                            "Found source port: %s" % source_port.name
                        )
                    if (
                        connect_with_port is not None
                        and connect_with_port_pattern is not None
                    ):
                        if not fnmatch.fnmatch(
                            connect_with_port, connect_with_port_pattern
                        ):
                            pc_logging.error(
                                "The determined source port does not match the pattern: %s"
                                % connect_with_port_pattern
                            )

                    # If the target port is determined, then use it
                    if not connect_to_port is None and target_port is None:
                        target_port = target_part.with_ports.get_ports()[
                            connect_to_port
                        ]
                        pc_logging.debug(
                            "Found target port: %s" % target_port.name
                        )
                    if (
                        connect_to_port is not None
                        and connect_to_port_pattern is not None
                    ):
                        if not fnmatch.fnmatch(
                            connect_to_port, connect_to_port_pattern
                        ):
                            pc_logging.error(
                                "The determined target port does not match the pattern: %s"
                                % connect_to_port_pattern
                            )

                    # TODO(clairbee): before the next step, deduce the interface
                    #                 based on the port name if the interface is missing

                    # Now calculate offsets based on params.
                    # This requires an interface object to be present, as that's where the params are defined.
                    # If the source interface params are passed, calculate the offsets
                    if (
                        source_iface_obj is not None
                        and connect_with_params is not None
                    ):
                        pc_logging.debug("Source params are found")
                        for (
                            param_name,
                            param_value,
                        ) in connect_with_params.items():
                            pc_logging.debug("Source param: %s" % param_name)
                            param = source_iface_obj.params.get(
                                param_name, None
                            )
                            pc_logging.debug("Source param: %s" % param)
                            if param is not None:
                                offsets = param.get_offsets(param_value)
                                pc_logging.debug("Source offsets: %s" % offsets)
                                source_offsets.extend(offsets)

                    # If the target interface params are passed, calculate the offsets
                    if (
                        target_iface_obj is not None
                        and connect_to_params is not None
                    ):
                        pc_logging.debug("Target params are found")
                        for (
                            param_name,
                            param_value,
                        ) in connect_to_params.items():
                            pc_logging.debug("Target param: %s" % param_name)
                            pc_logging.debug(
                                "Target info: %s" % target_iface_obj.info()
                            )
                            param = target_iface_obj.params.get(
                                param_name, None
                            )
                            pc_logging.debug("Target param: %s" % param)
                            if param is not None:
                                offsets = param.get_offsets(param_value)
                                pc_logging.debug("Target offsets: %s" % offsets)
                                target_offsets.extend(offsets)

                    if (source_port is None and target_port is not None) or (
                        source_port is not None and target_port is None
                    ):
                        # One of the parts may have no parts declared.
                        # TODO(clairbee): Do we need to support this?
                        pc_logging.warning(
                            "Peer port auto-detection has failed: %s" % name
                        )

                    turn_around = b3d.Location(
                        (0, 0, 0),
                        (0.71, 0.71, 0),
                        180,
                    ).wrapped.Transformation()

                    if source_port is not None and target_port is not None:
                        pc_logging.debug(
                            "Connected %s of %s to %s of %s"
                            % (
                                connect_with_port,
                                name,
                                connect_to_port,
                                connect_to_name,
                            )
                        )

                        trsf = target_part_location.wrapped.Transformation()
                        trsf.Multiply(
                            target_port.location.wrapped.Transformation()
                        )
                        trsf.Multiply(turn_around)
                        for target_offset in target_offsets:
                            pc_logging.debug(
                                "Target offset: %s" % target_offset
                            )
                            trsf.Multiply(target_offset)
                        for source_offset in source_offsets:
                            trsf.Multiply(source_offset)
                        trsf.Multiply(
                            source_port.location.wrapped.Transformation().Inverted()
                        )
                        location = b3d.Location(trsf)
                    elif source_port is None and target_port is not None:
                        pc_logging.debug(
                            "Connected %s to %s of %s"
                            % (
                                name,
                                connect_to_port,
                                connect_to_name,
                            )
                        )

                        trsf = target_part_location.wrapped.Transformation()
                        trsf.Multiply(
                            target_port.location.wrapped.Transformation()
                        )
                        trsf.Multiply(turn_around)
                        for target_offset in target_offsets:
                            trsf.Multiply(target_offset)
                        location = b3d.Location(trsf)
                    elif source_port is not None and target_port is None:
                        pc_logging.debug(
                            "Connected %s of %s to %s"
                            % (connect_with_port, name, connect_to_name)
                        )
                        trsf = target_part_location.wrapped.Transformation()
                        trsf.Multiply(turn_around)
                        for source_offset in source_offsets:
                            trsf.Multiply(source_offset)
                        trsf.Multiply(
                            source_port.location.wrapped.Transformation().Inverted()
                        )
                        location = b3d.Location(trsf)
                    elif source_port is None and target_port is None:
                        pc_logging.debug(
                            "Connected %s to %s" % (name, connect_to_name)
                        )
                        trsf = target_part_location.wrapped.Transformation()
                        trsf.Multiply(turn_around)
                        location = b3d.Location(trsf)
                    else:
                        pc_logging.error("Not enough data to connect %s" % name)
                        location = b3d.Location((0, 0, 0), (0, 0, 1), 0)

        if not item is None:
            return [item, name, location]
        else:
            return None



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-04-20
#
# Licensed under Apache License, Version 2.0.
#

import typing

from .sketch import Sketch
from .shape_factory import ShapeFactory


class SketchFactory(ShapeFactory):
    # TODO(clairbee): Make the next line work for part_factory_file only
    path: typing.Optional[str] = None
    sketch: Sketch
    name: str
    orig_name: str

    def __init__(
        self,
        ctx,
        source_project,
        target_project,
        config: object,
    ):
        super().__init__(ctx, source_project, config)
        self.target_project = target_project
        self.name = config["name"]
        self.orig_name = config["orig_name"]

    def _create_sketch(self, config: object) -> Sketch:
        sketch = Sketch(config)
        sketch.project_name = (
            self.target_project.name
        )  # TODO(clairbee): pass it via the constructor
        # TODO(clairbee): Make the next line work for sketch_factory_file only
        sketch.instantiate = lambda sketch_self: self.instantiate(sketch_self)
        sketch.info = lambda: self.info(sketch)
        sketch.with_ports = self.with_ports
        return sketch

    def _create(self, config: object):
        self.sketch = self._create_sketch(config)
        if self.path:
            self.sketch.path = self.path

        self.target_project.sketches[self.name] = self.sketch
        self.ctx.stats_sketches += 1



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-09-07
#
# Licensed under Apache License, Version 2.0.
#

import tempfile

from .provider_request_avail import ProviderRequestAvail
from .provider_request_order import ProviderRequestOrder
from .provider_request_quote import ProviderRequestQuote
from .provider_factory_python import ProviderFactoryPython
from .provider_data_cart import *
from . import logging as pc_logging

# from . import sync_threads as pc_thread


class ProviderFactoryStore(ProviderFactoryPython):
    def __init__(self, ctx, source_project, target_project, config):
        with pc_logging.Action(
            "InitStore", target_project.name, config["name"]
        ):
            super().__init__(
                ctx,
                source_project,
                target_project,
                config,
            )
            # Complement the config object here if necessary

            self._create(config)
            self.provider.is_part_available = self.is_part_available
            self.provider.load = self.load
            self.provider.query_quote = self.query_quote
            self.provider.query_order = self.query_order

    async def is_part_available(self, cart_item: ProviderCartItem):
        request = ProviderRequestAvail(
            self.name,
            cart_item.vendor,
            cart_item.sku,
            cart_item.count_per_sku,
            cart_item.count,
        )
        availability = await self.query_script(
            self.provider,
            "avail",
            request.compose(),
        )
        return (
            availability["available"]
            if availability and "available" in availability
            else False
        )

    async def load(self, cart_item: ProviderCartItem):
        """No-op. No CAD binary to be loded."""
        pass

    async def query_quote(self, request: ProviderRequestQuote):
        # TODO(clairbee): does it make sense to run this in a separate thread?
        # return await pc_thread.run_async(
        #   self.query_script, self.provider, "quote", request.compose(),
        # )
        return await self.query_script(
            self.provider, "quote", request.compose()
        )

    async def query_order(self, request: ProviderRequestOrder):
        # TODO(clairbee): does it make sense to run this in a separate thread?
        # return await pc_thread.run_async(
        #   self.query_script, self.provider, "order", request.compose(),
        # )
        return await self.query_script(
            self.provider, "order", request.compose()
        )



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-02-04
#
# Licensed under Apache License, Version 2.0.

import atexit
import logging
from logging.handlers import QueueHandler, QueueListener
import queue
import sys
import time
from typing import Any
import threading

from .logging import ops, error

COLOR_DEBUG = "\033[94m"
COLOR_INFO = "\033[92m\033[1m"
COLOR_WARN = "\033[93m"
COLOR_ERROR = "\033[91m"
COLOR_CRIT = "\033[91m\033[1m"
COLOR_NONE = "\033[0m"

COLOR_ACTION = "\033[1m"

WRAP = "\033[?7h"
NO_WRAP = "\033[?7l"


def ansi_process_start(op: str, package: str, item: str = None):
    logging.getLogger("partcad").critical(
        "process_start: %s: %s: %s" % (op, package, item),
        extra={
            "pc_event": "process_start",
            "op": op,
            "package": package,
            "item": item,
        },
    )


def ansi_process_end(op: str, package: str, item: str = None):
    logging.getLogger("partcad").critical(
        "process_end: %s: %s: %s" % (op, package, item),
        extra={
            "pc_event": "process_end",
            "op": op,
            "package": package,
            "item": item,
        },
    )


def ansi_action_start(op: str, package: str, item: str = None):
    logging.getLogger("partcad").critical(
        "action_start: %s: %s: %s" % (op, package, item),
        extra={
            "pc_event": "action_start",
            "op": op,
            "package": package,
            "item": item,
        },
    )


def ansi_action_end(op: str, package: str, item: str = None):
    logging.getLogger("partcad").critical(
        "action_end: %s: %s: %s" % (op, package, item),
        extra={
            "pc_event": "action_end",
            "op": op,
            "package": package,
            "item": item,
        },
    )


class AnsiTerminalProgressHandler(logging.Handler):
    MAX_LINES = 8

    def __init__(self, stream=sys.stdout) -> None:
        super().__init__()
        self.thread_lock = threading.Lock()
        if not stream is None:
            self.stream = stream
        else:
            self.stream = sys.stdout

        self.thread = None

        self.process = None
        self.process_target = None
        self.process_start = None
        self.footer_size = 0

        self.actions = {}
        self.actions_running = 0
        self.actions_total = 0

        self.last_output = 0.0

    def clear_footer(self):
        return "\u001b[1A\u001b[2K" * self.footer_size

    def run_thread(self):
        while not self.process is None:
            time.sleep(0.25)
            if time.time() - self.last_output > 0.5:
                self.emit(None)

    def emit(self, record):
        now = time.time()

        # output accumulates the string that will be emitted before 'return'
        output = ""

        # protect the status data store in 'self' from other threads
        self.thread_lock.acquire()

        if not record is None:
            ignore_message = False
            if hasattr(record, "pc_event"):
                if record.pc_event == "process_start":
                    self.process = record.op
                    self.process_target = record.package
                    if record.item is not None:
                        self.process_target += ":" + record.item
                    self.process_start = time.time()
                    self.actions_total = 0

                    if not self.thread is None:
                        self.thread_lock.release()
                        raise Exception("nested processes")
                    self.thread = threading.Thread(
                        target=self.run_thread,
                        name="ansi-terminal-update-" + str(now),
                    )
                    self.thread_lock.release()
                    self.thread.start()
                    return
                elif record.pc_event == "process_end":
                    self.process = None  # This also signals the thread to stop
                    output += self.clear_footer()
                    self.footer_size = 0

                    self.last_output = 0.0
                    th = self.thread
                    self.thread = None
                    self.thread_lock.release()

                    th.join()

                    if len(output) > 0:
                        self.stream.write(output)
                        self.stream.flush()
                    return

                elif record.pc_event == "action_start":
                    action = {"op": record.op}
                    target = record.package
                    if record.item is not None:
                        target += ":" + record.item
                    action["target"] = target
                    action["start"] = time.time()
                    self.actions[record.op + "-" + target] = action

                    self.actions_running += 1
                    self.actions_total += 1
                elif record.pc_event == "action_end":
                    target = record.package
                    if record.item is not None:
                        target += ":" + record.item
                    action_key = record.op + "-" + target
                    if action_key in self.actions:
                        del self.actions[record.op + "-" + target]
                    else:
                        error(
                            "action_key not found: %s: among %s"
                            % (action_key, str(self.actions.keys()))
                        )

                    self.actions_running -= 1
                ignore_message = True

        output += self.clear_footer()

        if not record is None:
            if not ignore_message:
                if record.levelno == logging.DEBUG:
                    output += COLOR_DEBUG + "DEBUG:" + COLOR_NONE + " "
                elif record.levelno == logging.INFO:
                    output += COLOR_INFO + "INFO: " + COLOR_NONE + " "
                elif record.levelno == logging.WARN:
                    output += COLOR_WARN + "WARN:" + COLOR_NONE + " "
                elif record.levelno == logging.ERROR:
                    output += COLOR_ERROR + "ERROR:" + COLOR_NONE + " "
                elif record.levelno == logging.CRITICAL:
                    output += COLOR_CRIT + "CRIT:" + COLOR_NONE + " "

                msg = self.format(record)
                output += "%s\n" % msg

        if not self.process is None:
            seconds = int(now - self.process_start)

            output += NO_WRAP
            output += "%s[ %d / %d ]%s %s%s%s %s; %ds\n" % (
                COLOR_INFO,
                self.actions_running,
                self.actions_total,
                COLOR_NONE,
                COLOR_ACTION,
                self.process,
                COLOR_NONE,
                self.process_target,
                seconds,
            )
            self.footer_size = 1

            sorted_actions = sorted(
                self.actions.values(), key=lambda a: a["start"]
            )
            sorted_actions = sorted_actions[: self.MAX_LINES]
            for action in sorted_actions:
                output += "%s\t[%s]%s %s [%ds]\n" % (
                    COLOR_ACTION,
                    action["op"],
                    COLOR_NONE,
                    action["target"],
                    now - action["start"],
                )
                self.footer_size += 1
            output += WRAP
        else:
            self.footer_size = 0

        self.last_output = now
        self.thread_lock.release()
        if len(output) > 0:
            self.stream.write(output)
            self.stream.flush()

        return


listener = None
queue_handler: AnsiTerminalProgressHandler = None
old_handlers = []
log_queue = None


def fini():
    global listener
    global queue_handler

    if queue_handler is not None:
        logger = logging.getLogger("partcad")
        for handler in logger.handlers:
            if handler == queue_handler:
                logger.removeHandler(handler)
        for handler in old_handlers:
            logger.addHandler(handler)
        queue_handler = None

    if listener is not None:
        atexit.unregister(fini)
        listener.stop()
        listener = None


def init(stream=None):
    global listener
    global queue_handler
    global log_queue
    global ops

    if log_queue is None:
        log_queue = queue.Queue(-1)

    if queue_handler is None:
        queue_handler = QueueHandler(log_queue)

        logger = logging.getLogger("partcad")
        handlers = logger.handlers
        for handler in handlers:
            old_handlers.append(handler)
            logger.removeHandler(handler)
        logger.addHandler(queue_handler)

    if listener is None:
        output_handler = AnsiTerminalProgressHandler(stream=stream)
        listener = QueueListener(log_queue, output_handler)
        listener.start()
        atexit.register(fini)

    ops.process_start = ansi_process_start
    ops.process_end = ansi_process_end
    ops.action_start = ansi_action_start
    ops.action_end = ansi_action_end



#
# OpenVMP, 2023-2024
#
# Author: Roman Kuzmenko
# Created: 2024-04-08
#
# Licensed under Apache License, Version 2.0.
#

from enum import Enum
from ocp_tessellate.ocp_utils import (
    is_vector,
    is_topods_shape,
    is_topods_compound,
    is_cadquery,
    is_cadquery_assembly,
    is_cadquery_sketch,
    is_build123d,
    is_toploc_location,
    is_wrapped,
)

saved_show_object = None


class Camera(Enum):
    """Camera reset modes"""

    RESET = "reset"
    CENTER = "center"
    KEEP = "keep"


class Collapse(Enum):
    """Collapse modes for the CAD navigation tree"""

    NONE = 0
    LEAVES = 1
    ALL = 2
    ROOT = 3


def show_object(
    cad_obj,
    name=None,
    options=None,
    parent=None,
    clear=False,
    port=None,
    progress="-+c",  # Ignore it, we don't want any output
    glass=None,
    tools=None,
    measure_tools=None,
    tree_width=None,
    axes=None,
    axes0=None,
    grid=None,
    ortho=None,
    transparent=None,
    default_opacity=None,
    black_edges=None,
    orbit_control=None,
    collapse=None,
    ticks=None,
    center_grid=None,
    up=None,
    zoom=None,
    position=None,
    quaternion=None,
    target=None,
    reset_camera=None,
    clip_slider_0=None,
    clip_slider_1=None,
    clip_slider_2=None,
    clip_normal_0=None,
    clip_normal_1=None,
    clip_normal_2=None,
    clip_intersection=None,
    clip_planes=None,
    clip_object_colors=None,
    pan_speed=None,
    rotate_speed=None,
    zoom_speed=None,
    deviation=None,
    angular_tolerance=None,
    edge_accuracy=None,
    default_color=None,
    default_facecolor=None,
    default_thickedgecolor=None,
    default_vertexcolor=None,
    default_edgecolor=None,
    ambient_intensity=None,
    metalness=None,
    roughness=None,
    direct_intensity=None,
    render_edges=None,
    render_normals=None,
    render_mates=None,
    render_joints=None,
    parallel=None,
    show_parent=None,
    show_sketch_local=None,
    helper_scale=None,
    mate_scale=None,  # DEPRECATED
    debug=None,
    timeit=None,
):
    if options is None:
        options = {}
    options.update(
        {
            "name": name,
            "parent": parent,
            "clear": clear,
            "port": port,
            # "progress": progress,
            "glass": glass,
            "tools": tools,
            "measure_tools": measure_tools,
            "tree_width": tree_width,
            "axes": axes,
            "axes0": axes0,
            "grid": grid,
            "ortho": ortho,
            "transparent": transparent,
            "default_opacity": default_opacity,
            "black_edges": black_edges,
            "orbit_control": orbit_control,
            "collapse": collapse,
            "ticks": ticks,
            "center_grid": center_grid,
            "up": up,
            "zoom": zoom,
            "position": position,
            "quaternion": quaternion,
            "target": target,
            "reset_camera": reset_camera,
            "clip_slider_0": clip_slider_0,
            "clip_slider_1": clip_slider_1,
            "clip_slider_2": clip_slider_2,
            "clip_normal_0": clip_normal_0,
            "clip_normal_1": clip_normal_1,
            "clip_normal_2": clip_normal_2,
            "clip_intersection": clip_intersection,
            "clip_planes": clip_planes,
            "clip_object_colors": clip_object_colors,
            "pan_speed": pan_speed,
            "rotate_speed": rotate_speed,
            "zoom_speed": zoom_speed,
            "deviation": deviation,
            "angular_tolerance": angular_tolerance,
            "edge_accuracy": edge_accuracy,
            "default_color": default_color,
            "default_facecolor": default_facecolor,
            "default_thickedgecolor": default_thickedgecolor,
            "default_vertexcolor": default_vertexcolor,
            "default_edgecolor": default_edgecolor,
            "ambient_intensity": ambient_intensity,
            "metalness": metalness,
            "roughness": roughness,
            "direct_intensity": direct_intensity,
            "render_edges": render_edges,
            "render_normals": render_normals,
            "render_mates": render_mates,
            "render_joints": render_joints,
            "parallel": parallel,
            "show_parent": show_parent,
            "show_sketch_local": show_sketch_local,
            "helper_scale": helper_scale,
            "mate_scale": mate_scale,  # DEPRECATED
            "debug": debug,
            "timeit": timeit,
        }
    )
    saved_show_object(
        cad_obj,
        options=options,
    )


def show(
    *args,
    names=None,
    colors=None,
    alphas=None,
    port=None,
    progress="-+c",  # Ignore it, we don't want any output
    glass=None,
    tools=None,
    measure_tools=None,
    tree_width=None,
    axes=None,
    axes0=None,
    grid=None,
    ortho=None,
    transparent=None,
    default_opacity=None,
    black_edges=None,
    orbit_control=None,
    collapse=None,
    explode=None,
    ticks=None,
    center_grid=None,
    up=None,
    zoom=None,
    position=None,
    quaternion=None,
    target=None,
    reset_camera=None,
    clip_slider_0=None,
    clip_slider_1=None,
    clip_slider_2=None,
    clip_normal_0=None,
    clip_normal_1=None,
    clip_normal_2=None,
    clip_intersection=None,
    clip_planes=None,
    clip_object_colors=None,
    pan_speed=None,
    rotate_speed=None,
    zoom_speed=None,
    deviation=None,
    angular_tolerance=None,
    edge_accuracy=None,
    default_color=None,
    default_edgecolor=None,
    default_facecolor=None,
    default_thickedgecolor=None,
    default_vertexcolor=None,
    ambient_intensity=None,
    direct_intensity=None,
    metalness=None,
    roughness=None,
    render_edges=None,
    render_normals=None,
    render_mates=None,
    render_joints=None,
    show_parent=None,
    show_sketch_local=None,
    parallel=None,
    helper_scale=None,
    mate_scale=None,  # DEPRECATED
    debug=None,
    timeit=None,
    _force_in_debug=False,
):
    for idx, cad_obj in enumerate(args):
        saved_show_object(
            cad_obj,
            options={
                "name": names[idx] if names is not None else None,
                "color": colors[idx] if colors is not None else None,
                "alpha": alphas[idx] if alphas is not None else None,
                "port": port,
                # "progress": progress,
                "glass": glass,
                "tools": tools,
                "measure_tools": measure_tools,
                "tree_width": tree_width,
                "axes": axes,
                "axes0": axes0,
                "grid": grid,
                "ortho": ortho,
                "transparent": transparent,
                "default_opacity": default_opacity,
                "black_edges": black_edges,
                "orbit_control": orbit_control,
                "collapse": collapse,
                "ticks": ticks,
                "center_grid": center_grid,
                "up": up,
                "zoom": zoom,
                "position": position,
                "quaternion": quaternion,
                "target": target,
                "reset_camera": reset_camera,
                "clip_slider_0": clip_slider_0,
                "clip_slider_1": clip_slider_1,
                "clip_slider_2": clip_slider_2,
                "clip_normal_0": clip_normal_0,
                "clip_normal_1": clip_normal_1,
                "clip_normal_2": clip_normal_2,
                "clip_intersection": clip_intersection,
                "clip_planes": clip_planes,
                "clip_object_colors": clip_object_colors,
                "pan_speed": pan_speed,
                "rotate_speed": rotate_speed,
                "zoom_speed": zoom_speed,
                "deviation": deviation,
                "angular_tolerance": angular_tolerance,
                "edge_accuracy": edge_accuracy,
                "default_color": default_color,
                "default_facecolor": default_facecolor,
                "default_thickedgecolor": default_thickedgecolor,
                "default_vertexcolor": default_vertexcolor,
                "default_edgecolor": default_edgecolor,
                "ambient_intensity": ambient_intensity,
                "metalness": metalness,
                "roughness": roughness,
                "direct_intensity": direct_intensity,
                "render_edges": render_edges,
                "render_normals": render_normals,
                "render_mates": render_mates,
                "render_joints": render_joints,
                "parallel": parallel,
                "show_parent": show_parent,
                "show_sketch_local": show_sketch_local,
                "helper_scale": helper_scale,
                "mate_scale": mate_scale,  # DEPRECATED
                "debug": debug,
                "timeit": timeit,
            },
        )


def show_clear():
    pass


def show_all(
    variables=None, exclude=None, classes=None, _visual_debug=False, **kwargs
):
    import inspect
    import re

    if variables is None:
        cf = inspect.currentframe()
        variables = cf.f_back.f_locals

    if exclude is None:
        exclude = []

    objects = []
    names = []
    for name, obj in variables.items():
        if (
            isinstance(obj, type)
            or name in ["_", "__", "___"]
            or name.startswith("__")
            or re.search("_\\d+", name) is not None
        ):
            continue  # ignore classes and jupyter variables
        if hasattr(obj, "area") and obj.area > 1e99:  # inifinite face
            print(f"infinite face {name} skipped")
            continue

        if name not in exclude:
            if (
                hasattr(obj, "_obj")
                and obj._obj is None  # pylint: disable=protected-access
            ):
                continue

            if hasattr(obj, "locations") and hasattr(obj, "local_locations"):
                obj = obj.locations

            if hasattr(obj, "local_coord_system"):
                obj = obj.location

            if (
                (
                    hasattr(obj, "wrapped")
                    and (
                        is_topods_shape(obj.wrapped)
                        or is_topods_compound(obj.wrapped)
                        or is_toploc_location(obj.wrapped)
                    )
                )
                or is_vector(obj)  # Vector
                or is_cadquery(obj)
                or is_build123d(obj)
                or is_cadquery_assembly(obj)
                or (
                    hasattr(obj, "wrapped")
                    and hasattr(obj, "position")
                    and hasattr(obj, "direction")
                )
            ):
                objects.append(obj)
                names.append(name)

    if len(objects) > 0:
        show(
            *objects,
            names=names,
            collapse=Collapse.ROOT,
            _force_in_debug=_visual_debug,
            **kwargs,
        )
    else:
        show_clear()


def set_defaults(**kwargs):
    pass



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-12-30
#
# Licensed under Apache License, Version 2.0.
#

# This script is executed within the python sandbox environment (python runtime)
# to invoke `cadquery` scripts.

import os
import sys

import cadquery as cq

sys.path.append(os.path.dirname(__file__))
import wrapper_common


def process(path, request):
    shape = cq.importers.importStep(path).val().wrapped

    return {
        "success": True,
        "exception": None,
        "shape": shape,
    }


path, request = wrapper_common.handle_input()

# Call CadQuery
model = process(path, request)

wrapper_common.handle_output(model)



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-09-30
#
# Licensed under Apache License, Version 2.0.
#

import typing

from .assembly import Assembly
from .shape_factory import ShapeFactory


class AssemblyFactory(ShapeFactory):
    # TODO(clairbee): Make the next line work for assembly_factory_file only
    path: typing.Optional[str] = None
    assembly: Assembly

    def __init__(
        self, ctx, source_project, target_project, config, extension=""
    ):
        super().__init__(ctx, source_project, config)
        self.name = config["name"]
        self.orig_name = config["orig_name"]

    def _create(self, config):
        self.assembly = Assembly(config)
        self.assembly.project_name = (
            self.project.name
        )  # TODO(clairbee): pass it via the constructor
        # TODO(clairbee): Make the next line work for assembly_factory_file only
        if self.path:
            self.assembly.path = self.path
        self.project.assemblies[self.name] = self.assembly

        self.assembly.instantiate = lambda assembly_self: self.instantiate(
            assembly_self
        )
        self.assembly.info = lambda: self.info(self.assembly)
        self.assembly.with_ports = self.with_ports

        self.ctx.stats_assemblies += 1



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-04-20
#
# Licensed under Apache License, Version 2.0.
#

import base64
import os
import pickle
import sys

from OCP.gp import gp_Ax1
from OCP.TopoDS import (
    TopoDS_Builder,
    TopoDS_Compound,
    TopoDS_Edge,
    TopoDS_Wire,
    TopoDS_Face,
)
from OCP.TopLoc import TopLoc_Location

from .sketch_factory_python import SketchFactoryPython
from . import wrapper
from . import logging as pc_logging

sys.path.append(os.path.join(os.path.dirname(__file__), "wrappers"))
from cq_serialize import register as register_cq_helper


class SketchFactoryBuild123d(SketchFactoryPython):
    def __init__(
        self, ctx, source_project, target_project, config, can_create=False
    ):
        python_version = source_project.python_version
        if python_version is None:
            # Stay one step ahead of the minimum required Python version
            python_version = "3.10"
        if python_version == "3.12" or python_version == "3.11":
            pc_logging.debug(
                "Downgrading Python version to 3.10 to avoid compatibility issues with build123d"
            )
            python_version = "3.10"
        with pc_logging.Action(
            "InitBuild123d", target_project.name, config["name"]
        ):
            super().__init__(
                ctx,
                source_project,
                target_project,
                config,
                can_create=can_create,
                python_version=python_version,
            )
            # Complement the config object here if necessary
            self._create(config)

    async def instantiate(self, sketch):
        await super().instantiate(sketch)

        with pc_logging.Action("Build123d", sketch.project_name, sketch.name):
            if (
                not os.path.exists(sketch.path)
                or os.path.getsize(sketch.path) == 0
            ):
                pc_logging.error(
                    "build123d script is empty or does not exist: %s"
                    % sketch.path
                )
                return None

            # Finish initialization of PythonRuntime
            # which was too expensive to do in the constructor
            await self.prepare_python()

            # Get the path to the wrapper script
            # which needs to be executed
            wrapper_path = wrapper.get("build123d.py")

            # Build the request
            request = {"build_parameters": {}}
            if "parameters" in self.config:
                for param_name, param in self.config["parameters"].items():
                    request["build_parameters"][param_name] = param["default"]
            patch = {}
            if "show" in self.config:
                patch["\\Z"] = "\nshow(%s)\n" % self.config["show"]
            if "showObject" in self.config:
                patch["\\Z"] = "\nshow_object(%s)\n" % self.config["show"]
            if "patch" in self.config:
                patch.update(self.config["patch"])
            request["patch"] = patch

            # Serialize the request
            register_cq_helper()
            picklestring = pickle.dumps(request)
            request_serialized = base64.b64encode(picklestring).decode()

            await self.runtime.ensure("ocp-tessellate")
            await self.runtime.ensure("cadquery")
            await self.runtime.ensure("numpy==1.24.1")
            await self.runtime.ensure("numpy-quaternion==2023.0.4")
            await self.runtime.ensure("nptyping==1.24.1")
            await self.runtime.ensure("typing_extensions>=4.6.0,<5")
            await self.runtime.ensure("build123d")
            cwd = self.project.config_dir
            if self.cwd is not None:
                cwd = os.path.join(self.project.config_dir, self.cwd)
            response_serialized, errors = await self.runtime.run(
                [
                    wrapper_path,
                    os.path.abspath(sketch.path),
                    os.path.abspath(cwd),
                ],
                request_serialized,
            )
            if len(errors) > 0:
                error_lines = errors.split("\n")
                for error_line in error_lines:
                    error_line = error_line.strip()
                    if not error_line:
                        continue
                    # TODO(clairbee): Move the sketch name concatenation to where the logging happens
                    # part.error("%s: %s" % (sketch.name, error_line))
                    sketch.error(error_line)

            try:
                # pc_logging.error("Response: %s" % response_serialized)
                response = base64.b64decode(response_serialized)
                register_cq_helper()
                result = pickle.loads(response)
            except Exception as e:
                sketch.error(
                    "Exception while deserializing %s: %s" % (sketch.name, e)
                )
                return None

            if not result["success"]:
                sketch.error("%s: %s" % (sketch.name, result["exception"]))
                return None

            self.ctx.stats_sketches_instantiated += 1
            sketch.components = []

            if result["shapes"] is None:
                return None
            if len(result["shapes"]) == 0:
                return None

            builder = TopoDS_Builder()
            compound = TopoDS_Compound()
            builder.MakeCompound(compound)

            def process(shapes, components_list):
                for shape in shapes:
                    # pc_logging.info("Returned: %s" % type(shape))
                    try:
                        if shape is None or isinstance(shape, str):
                            # pc_logging.info("String: %s" % shape)
                            continue

                        if isinstance(shape, list):
                            child_component_list = list()
                            process(shape, child_component_list)
                            components_list.append(child_component_list)
                            continue

                        # TODO(clairbee): add support for the below types
                        if isinstance(shape, TopLoc_Location) or isinstance(
                            shape, gp_Ax1
                        ):
                            continue

                        if (
                            isinstance(shape, TopoDS_Edge)
                            or isinstance(shape, TopoDS_Wire)
                            or isinstance(shape, TopoDS_Face)
                        ):
                            builder.Add(compound, shape)
                            components_list.append(shape)
                        elif False:
                            # TODO(clairbee) Add all metadata types here
                            components_list.append(shape)
                        else:
                            pc_logging.error(
                                "Unsupported shape type: %s" % type(shape)
                            )
                    except Exception as e:
                        pc_logging.error(
                            "Error adding shape to compound: %s" % e
                        )

            process(result["shapes"], sketch.components)

            return compound



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-08-19
#
# Licensed under Apache License, Version 2.0.
#

from git import Repo
import hashlib
import os
import pathlib
import threading
import time

from . import project_factory as pf
from . import logging as pc_logging
from .user_config import user_config

global_cache_lock = threading.Lock()
cache_locks = {}


def get_cache_lock(hash):
    global global_cache_lock
    global_cache_lock.acquire()
    if hash not in cache_locks:
        cache_locks[hash] = threading.Lock()
    lock = cache_locks[hash]
    global_cache_lock.release()
    return lock


class GitImportConfiguration:
    def __init__(self):
        self.import_config_url = self.config_obj.get("url")
        self.import_revision = self.config_obj.get("revision")
        self.import_rel_path = self.config_obj.get("relPath")


class ProjectFactoryGit(pf.ProjectFactory, GitImportConfiguration):
    def __init__(self, ctx, parent, config):
        pf.ProjectFactory.__init__(self, ctx, parent, config)
        GitImportConfiguration.__init__(self)

        self.path = self._clone_or_update_repo(self.import_config_url)

        # Complement the config object here if necessary
        self._create(config)

        # TODO(clairbee): actually fill in the self.project object here

        self._save()

    def _clone_or_update_repo(self, repo_url, cache_dir=None):
        """
        Clones a Git repository to a local directory and keeps it up-to-date.

        Args:
          repo_url: URL of the Git repository to clone.
          cache_dir: Directory to store the cached copies of repositories (defaults to ".cache").

        Returns:
          Local path to the cloned repository.
        """

        if cache_dir is None:
            cache_dir = os.path.join(user_config.internal_state_dir, "git")

        # Generate a unique identifier for the repository based on its URL.
        repo_hash = hashlib.sha256(repo_url.encode()).hexdigest()
        if self.import_revision is not None:
            # Append the revision to the hash instead of using it as an input
            # to the hash function. This way we can navigate in the cache
            # a lot easier when there are multiple revisions of the same repo.
            display_rev = self.import_revision
            display_rev = display_rev.replace("/", "-slash-")
            if os.name == "nt":
                # On Windows, we need to replace backslashes as well.
                display_rev = display_rev.replace(os.path.sep, "-sep-")
            repo_hash += "-" + display_rev
        cache_path = os.path.join(cache_dir, repo_hash)
        cache_lock = get_cache_lock(repo_hash)

        guard_path = os.path.join(cache_path, ".partcad.git.cloned")

        with cache_lock:
            # Check if the repository is already cached.
            if os.path.exists(cache_path):
                # Update the repository if it is already cached.
                try:
                    before = None

                    # Try to open the existing repository and update it.
                    if self.import_revision is None:
                        # Import the default branch
                        if user_config.force_update or (
                            time.time() - os.path.getmtime(guard_path)
                            > 24 * 3600
                        ):
                            repo = Repo(cache_path)
                            origin = repo.remote("origin")
                            before = repo.active_branch.commit

                            # If there is more than 1 remote branch, we have to
                            # explicitly specify the branch to pull.
                            remote_head = origin.refs.HEAD
                            branch_name = remote_head.reference.name
                            short_branch_name = branch_name[
                                branch_name.find("/") + 1 :
                            ]
                            pc_logging.debug(
                                "Refreshing the GIT branch: %s"
                                % short_branch_name
                            )
                            origin.pull(short_branch_name)
                            pathlib.Path(guard_path).touch()
                    else:
                        # Import a specific revision
                        repo = Repo(cache_path)
                        origin = repo.remote("origin")
                        before = repo.active_branch.commit
                        if user_config.force_update or (
                            before != self.import_revision
                            or (
                                time.time() - os.path.getmtime(guard_path)
                                > 24 * 3600
                            )
                        ):
                            # Need to check for updates
                            origin.fetch()
                            repo.git.checkout(self.import_revision, force=True)
                            origin.pull()
                            pathlib.Path(guard_path).touch()

                    if not before is None:
                        # Update was performed
                        after = repo.active_branch.commit
                        if before != after:
                            pc_logging.info(
                                "Updated the GIT repo: %s"
                                % self.import_config_url
                            )
                except Exception as e:
                    pc_logging.error("Failed to update a repo: %s" % e)
                    # Fall back to using the previous copy
            else:
                # Clone the repository if it's not cached yet.
                try:
                    pc_logging.info(
                        "Cloning the GIT repo: %s" % self.import_config_url
                    )
                    repo = Repo.clone_from(repo_url, cache_path)
                    if not self.import_revision is None:
                        repo.git.checkout(self.import_revision, force=True)

                    if not os.path.exists(guard_path):
                        pathlib.Path(guard_path).touch()
                except Exception as e:
                    raise RuntimeError(f"Failed to clone a repo: {e}")

        if not self.import_rel_path is None:
            cache_path = os.path.join(cache_path, self.import_rel_path)

        return cache_path



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-03-23
#
# Licensed under Apache License, Version 2.0.
#

import fnmatch
import time

from .ai_google import AiGoogle
from .ai_openai import AiOpenAI
from .ai_ollama import AiOllama

from . import logging as pc_logging
from .user_config import user_config


supported_models = [
    "gpt-3.5-turbo",
    "gpt-4",
    "gpt-4-vision-preview",
    "gpt-4o",
    "gpt-4o-mini",
    "o1-*",
    "gemini-pro",
    "gemini-pro-vision",
    "gemini-1.5-pro",
    "gemini-1.5-flash",
    "llama3.*",
    "codellama*",
    "codegemma*",
    "gemma*",
    "deepseek-coder*",
    "codestral*",
]


class Ai(AiGoogle, AiOpenAI, AiOllama):
    def generate(
        self,
        action: str,
        package: str,
        item: str,
        prompt: str,
        config: dict[str, str],
        num_options: int = 1,
    ):
        with pc_logging.Action("Ai" + action, package, item):
            # Determine the model to use
            provider = config.get("provider", None)
            if (
                "model" in config
                and config["model"] is not None
                and config["model"] != ""
            ):
                model = config["model"]
            else:
                if provider is None:
                    if not user_config.openai_api_key is None:
                        provider = "openai"
                    else:
                        provider = "google"

                if provider == "google":
                    # if len(image_filenames) > 0:
                    #     model = "gemini-pro-vision"
                    # else:
                    #     model = "gemini-pro"
                    model = "gemini-1.5-pro"
                elif provider == "openai":
                    # if len(image_filenames) > 0:
                    #     model = "gpt-4o"
                    # else:
                    #     model = "gpt-4o"
                    model = "gpt-4o"
                elif provider == "ollama":
                    model = "llama3.1:70b"
                else:
                    error = "Provider %s is not supported" % provider
                    pc_logging.error(error)
                    return []

            # Generate the content
            is_supported = False
            for supported_model_pattern in supported_models:
                if fnmatch.fnmatch(model, supported_model_pattern):
                    is_supported = True
                    break
            if not is_supported:
                error = "Model %s is not supported" % model
                pc_logging.error(error)
                return []

            result = []
            if provider == "google":
                try:
                    result = self.generate_google(
                        model,
                        prompt,
                        config,
                        num_options,
                    )
                except Exception as e:
                    pc_logging.error(
                        "Failed to generate with Google: %s" % str(e)
                    )
                    time.sleep(1)  # Safeguard against exceeding quota

            elif provider == "openai":
                try:
                    result = self.generate_openai(
                        model,
                        prompt,
                        config,
                        num_options,
                    )
                except Exception as e:
                    pc_logging.error(
                        "Failed to generate with OpenAI: %s" % str(e)
                    )
                    time.sleep(1)  # Safeguard against exceeding quota

            elif provider == "ollama":
                try:
                    result = self.generate_ollama(
                        model,
                        prompt,
                        config,
                        num_options,
                    )
                except Exception as e:
                    pc_logging.error(
                        "Failed to generate with Ollama: %s" % str(e)
                    )

            else:
                pc_logging.error(
                    "Failed to associate the model %s with the provider" % model
                )

            if result is None:
                return []
            return result



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-03-16
#
# Licensed under Apache License, Version 2.0.
#

# This script is executed within a python runtime environment
# (no need for a sandbox) to speed up parallel rendering

import os
import sys

import build123d as b3d

sys.path.append(os.path.dirname(__file__))
import wrapper_common


def process(path, request):
    try:
        b3d_obj = b3d.Solid.make_box(1, 1, 1)
        b3d_obj.wrapped = request["wrapped"]

        viewport_origin = tuple(request["viewport_origin"])
        visible, hidden = b3d_obj.project_to_viewport(
            viewport_origin=viewport_origin
        )
        # visible = b3d_obj.project_to_viewport(
        #     viewport_origin=viewport_origin,
        #     ignore_hidden=True,
        # )[0]
        max_dimension = max(
            # *b3d.Compound(children=visible + hidden)
            *b3d.Compound(children=visible)
            .bounding_box()
            .size
        )
        if max_dimension == 0:
            max_dimension = 4
        scale = 512.0 / max_dimension
        exporter = b3d.ExportSVG(
            scale=scale,
            precision=10,
        )
        exporter.add_layer(
            "Visible",
            line_color=(64, 192, 64),
            line_weight=request["line_weight"],
        )
        # exporter.add_layer(
        #     "Hidden",
        #     line_color=(32, 64, 32),
        #     line_type=b3d.LineType.ISO_DOT,
        # )
        try:
            exporter.add_shape(visible, layer="Visible")
            # exporter.add_shape(hidden, layer="Hidden")
        except:
            pass
        exporter.write(path)

        return {
            "success": True,
            "exception": None,
        }
    except Exception as e:
        wrapper_common.handle_exception(e)
        return {
            "success": False,
            "exception": str(e.with_traceback(None)),
        }


path, request = wrapper_common.handle_input()

# Perform rendering
response = process(path, request)

wrapper_common.handle_output(response)



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-09-07
#
# Licensed under Apache License, Version 2.0.
#


class ProviderRequest:
    def compose(self) -> dict:
        # This method must be implemented in the child class
        raise NotImplementedError()



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-04-20
#
# Licensed under Apache License, Version 2.0.
#

from OCP.gp import (
    gp_Vec,
)
from OCP.BRepPrimAPI import (
    BRepPrimAPI_MakePrism,
)

from .part_factory import PartFactory
from .sketch import Sketch
from . import logging as pc_logging
from .utils import resolve_resource_path


class PartFactoryExtrude(PartFactory):
    depth: float
    source_project_name: str
    source_sketch_name: str
    source_sketch_spec: str
    sketch: Sketch

    def __init__(self, ctx, source_project, target_project, config):
        with pc_logging.Action(
            "IniExtrude", target_project.name, config["name"]
        ):
            super().__init__(
                ctx,
                source_project,
                target_project,
                config,
            )

            self.depth = float(config["depth"])

            self.source_sketch_name = config.get("sketch", "sketch")
            if "project" in config:
                self.source_project_name = config["project"]
                if (
                    self.source_project_name == "this"
                    or self.source_project_name == ""
                ):
                    self.source_project_name = source_project.name
            else:
                if ":" in self.source_sketch_name:
                    self.source_project_name, self.source_sketch_name = (
                        resolve_resource_path(
                            source_project.name,
                            self.source_sketch_name,
                        )
                    )
                else:
                    self.source_project_name = source_project.name
            self.source_sketch_spec = (
                self.source_project_name + ":" + self.source_sketch_name
            )

            self._create(config)

    async def instantiate(self, part):
        with pc_logging.Action("Extrude", part.project_name, part.name):
            shape = None
            try:
                self.sketch = self.project.ctx.get_sketch(
                    self.source_sketch_spec
                )

                maker = BRepPrimAPI_MakePrism(
                    await self.sketch.get_shape(),
                    gp_Vec(0.0, 0.0, self.depth),
                )
                maker.Build()
                shape = maker.Shape()
            except Exception as e:
                pc_logging.exception(
                    "Failed to create an extruded part: %s" % e
                )

            self.ctx.stats_parts_instantiated += 1

            return shape



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-01-22
#
# Licensed under Apache License, Version 2.0.
#

import typing

from . import part_factory as pf
from . import logging as pc_logging
from .utils import resolve_resource_path


class PartFactoryAlias(pf.PartFactory):
    source_part_name: str
    source_project_name: typing.Optional[str]
    source: str

    def __init__(self, ctx, source_project, target_project, config):
        with pc_logging.Action(
            "InitAlias", target_project.name, config["name"]
        ):
            super().__init__(ctx, source_project, target_project, config)
            # Complement the config object here if necessary
            self._create(config)

            if "source" in config:
                self.source_part_name = config["source"]
            else:
                self.source_assembly_name = config["name"]
                if not "project" in config:
                    raise Exception(
                        "Alias needs either the source part name or the source project name"
                    )

            if "project" in config:
                self.source_project_name = config["project"]
                if (
                    self.source_project_name == "this"
                    or self.source_project_name == ""
                ):
                    self.source_project_name = source_project.name
            else:
                if ":" in self.source_part_name:
                    self.source_project_name, self.source_part_name = (
                        resolve_resource_path(
                            source_project.name,
                            self.source_part_name,
                        )
                    )
                else:
                    self.source_project_name = source_project.name
            self.source = self.source_project_name + ":" + self.source_part_name
            config["source_resolved"] = self.source

            if self.source_project_name == target_project.name:
                self.part.desc = "Alias to %s" % self.source_part_name
            else:
                self.part.desc = "Alias to %s from %s" % (
                    self.source_part_name,
                    self.source_project_name,
                )

            pc_logging.debug("Initializing an alias to %s" % self.source)

    async def instantiate(self, part):
        with pc_logging.Action("Alias", part.project_name, part.name):

            source = self.ctx._get_part(self.source)
            shape = source.shape
            if not shape is None:
                return shape

            self.ctx.stats_parts_instantiated += 1

            if not source.path is None:
                part.path = source.path
            return await source.instantiate(part)



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-03-23
#
# Licensed under Apache License, Version 2.0.
#

import asyncio

import copy
import os
import tempfile
import threading

from .ai import Ai
from . import logging as pc_logging
from . import sync_threads as pc_thread
from .utils import total_size
from .user_config import user_config

DEFAULT_ALTERNATIVES_GEOMETRIC_MODELING = 3
DEFAULT_ALTERNATIVES_MODEL_GENERATION = 3
DEFAULT_INCREMENTAL_SCRIPT_CORRECTION = 2


class PartFactoryFeatureAi(Ai):
    """Used by all part factories that generate parts with GenAI."""

    LANG_PYTHON = "Python"

    part_type: str
    script_type: str
    prompt_suffix: str

    def __init__(self, config, part_type, script_type, prompt_suffix=""):
        self.part_type = part_type
        self.script_type = script_type
        self.prompt_suffix = prompt_suffix

        desc = config.get("desc", None)
        if desc is None:
            error = "%s: Prompt is not set" % config["name"]
            pc_logging.error(error)
            raise Exception(error)

        self.ai_config = copy.deepcopy(config)
        if "numGeometricModeling" in self.ai_config:
            self.num_geometric_modling = self.ai_config["numGeometricModeling"]
        else:
            self.num_geometric_modeling = (
                DEFAULT_ALTERNATIVES_GEOMETRIC_MODELING
            )
        if (
            user_config.max_geometric_modeling is not None
            and self.num_geometric_modeling > user_config.max_geometric_modeling
        ):
            self.num_geometric_modeling = user_config.max_geometric_modeling

        if "numModelGeneration" in self.ai_config:
            self.num_model_generation = self.ai_config[("numModelGeneration")]
        else:
            self.num_model_generation = DEFAULT_ALTERNATIVES_MODEL_GENERATION
        if (
            user_config.max_model_generation is not None
            and self.num_model_generation > user_config.max_model_generation
        ):
            self.num_model_generation = user_config.max_model_generation

        if "numScriptCorrection" in self.ai_config:
            self.num_script_correction = self.ai_config[("numScriptCorrection")]
        else:
            self.num_script_correction = DEFAULT_INCREMENTAL_SCRIPT_CORRECTION
        if (
            user_config.max_script_correction is not None
            and self.num_script_correction > user_config.max_script_correction
        ):
            self.num_script_correction = user_config.max_script_correction

        # Normalize the input configuration
        pc_logging.debug("AI configuration: %s" % self.ai_config)

        if (
            not "tokens" in self.ai_config
            or not isinstance(self.ai_config["tokens"], int)
            or self.ai_config["tokens"] == 0
        ):
            self.ai_config["tokens"] = 2048
            pc_logging.debug("Setting the default number of tokens: 2048")
        if not "images" in self.ai_config:
            self.ai_config["images"] = []
        # Use `temperature` and `top_p` values recommended for code generation
        # if no other preferences are set
        if not "temperature" in self.ai_config:
            self.ai_config["temperature"] = 0.2
        if not "top_p" in self.ai_config:
            self.ai_config["top_p"] = 0.1

    def on_init_ai(self):
        """This method must be executed at the very end of the part factory
        constructor to finalize the AI initialization. At the time of the call
        self.part and self.instantiate must be already defined."""
        self.part.generate = lambda path: self._create_file(path)
        self.part.change = lambda path, change=None: self._change_file(
            path, change
        )

        # If uncommented out, this makes the package initialization
        # unaccceptably slow
        # if (
        #     not os.path.exists(self.config["path"])
        #     or os.path.getsize(self.config["path"]) == 0
        # ):
        #     self._create_file()
        self.instantiate_orig = self.instantiate
        self.instantiate = self._instantiate_ai

    async def _instantiate_ai(self, part):
        """This is a wrapper for the instantiate method that ensures that
        the part is (re)generated before the instantiation."""
        if not os.path.exists(part.path) or os.path.getsize(part.path) == 0:
            try:
                self._create_file(part.path)
            except Exception as e:
                pc_logging.error(f"Failed to create the file: {e}")
                raise e

        return await self.instantiate_orig(part)

    def _create_file(self, path):
        """This method is called to (re)generate the part."""

        # CSG modeling
        modeling_options = []
        max_models = self.num_geometric_modeling
        max_tries = 2 * max_models

        # TODO(clairbee): Multiple AI calls are a good candidate to paralelize,
        #                 however it's useless without a paid subscription
        #                 with huge quotas. De-prioritized for now.
        #
        # def modeling_task():
        #     modeling_options.extend(self._csg_modeling())
        # threads = []
        # for _ in range(NUM_ALTERNATIVES_GEOMETRIC_MODELING):
        #     thread = threading.Thread(target=modeling_task)
        #     thread.start()
        #     threads.append(thread)
        # for thread in threads:
        #     thread.join()

        tries = 0
        while len(modeling_options) < max_models and tries < max_tries:
            modeling_options.extend(self._csg_modeling())
            pc_logging.info(
                "Generated %d CSG modeling candidates" % len(modeling_options)
            )
            tries += 1

        # For each remaining geometric modeling option,
        # generate a model and render an image
        script_candidates = []
        candidate_id = 0
        for modeling_option_idx, modeling_option in enumerate(modeling_options):
            pc_logging.debug(
                "Generated the geometric modeling candidate %d: %s"
                % (modeling_option_idx, modeling_option)
            )

            # Generate the models using the required language
            scripts = self._generate_script(modeling_option)

            for script in scripts:
                pc_logging.debug(
                    "Generated the script candidate %d: %s"
                    % (candidate_id, script)
                )

                # Validate the image by rendering it,
                # attempt to correct the script if rendering doesn't work
                image_filename, script = self._validate_and_fix(
                    script, candidate_id
                )
                # Check if the model was valid
                if image_filename is not None:
                    # Record the valid model and the image
                    script_candidates.append((image_filename, script))

                    # Once we generated a valid script and rendered the result,
                    # Attempt to change the script by comparing the result with
                    # the original request
                    changed_scripts = self._change_script(
                        modeling_option, script, image_filename
                    )
                    for changed_script in changed_scripts:
                        pc_logging.debug(
                            "Generated the changed script candidate %d: %s"
                            % (candidate_id, changed_script)
                        )

                        # Validate the image by rendering it,
                        # attempt to correct the script if rendering doesn't work
                        image_filename, changed_script = self._validate_and_fix(
                            changed_script, candidate_id
                        )
                        # Check if the model was valid
                        if image_filename is not None:
                            # Record the valid model and the image
                            script_candidates.append(
                                (image_filename, changed_script)
                            )

                candidate_id += 1

            pc_logging.info(
                "So far %d valid script candidates" % len(script_candidates)
            )

        if len(script_candidates) == 0:
            pc_logging.error(
                "No valid script generated. Try changing the prompt."
            )
            return

        if len(script_candidates) == 1:
            script = script_candidates[0][1]
        else:
            # Compare the images and select the best one
            script = self.select_best_image(script_candidates)

        if not script is None:
            f = open(path, "w")
            f.write(script)
            f.close()

    def _change_file(self, path, change=None):
        """This method is called to change the part."""

        script = open(path, "r").read()

        if change is not None:
            change = change.strip()
            if change == "":
                change = None

        image_filename, error_text = self._render_image(script, 0)
        if image_filename is None or error_text:
            pc_logging.error(
                "Failed to render the image for the script %s" % script
            )
            return

        script_candidates = []
        script_candidates.append((image_filename, script))

        # Attempt to change the script once more by comparing the result with
        # the original request
        changed_scripts = self._change_script(
            None, script, image_filename, change
        )
        for changed_script in changed_scripts:
            pc_logging.debug(
                "Generated the changed script: %s" % changed_script
            )

            # Validate the image by rendering it,
            # attempt to correct the script if rendering doesn't work
            new_image_filename, changed_script = self._validate_and_fix(
                changed_script,
                0,
            )
            # Check if the model was valid
            if new_image_filename is not None:
                # Record the valid model and the image
                script_candidates.append((new_image_filename, changed_script))

        if len(script_candidates) == 0:
            pc_logging.error(
                "No valid script generated. Try changing the prompt."
            )
            return

        if len(script_candidates) == 1:
            new_script = script_candidates[0][1]
        else:
            # Compare the images and select the best one
            new_script = self.select_best_image(
                script_candidates, change=change
            )

        if new_script == script:
            pc_logging.info("The script was not changed")
            return

        if not new_script is None:
            f = open(path, "w")
            f.write(new_script)
            f.close()

    def _csg_modeling(self):
        """This method generates CSG for the part."""

        prompt = (
            """You are an AI assistant to engineers modeling mechanical parts using constructive solid geometry.
Given a description of the part,
create a detailed sequence of instructions how to model this part using constructive solid geometry.
First, select an intuitive coordinate system for ease of placement for all primitives.
Then specify the minimum possible number of initial geometric primitives, their dimensions, location and orientation.
For simplicity, consider using large fillets on two edges of the cuboid to make a cylindrical end on a cuboid if needed, instead of adding a cylinder.
Also, for simplicity, consider using a chamfer to make a conical end on a cylinder if needed, insted of adding a cone.
Then specify how to locate and orient the primitives against each other.
Then specify what CSG operations to perform on sets of primitives
(unions, differences and intersections)
and on each of these primitives individually
(including but not limited to adding fillets, chamfers, paddings and cutting holes).

The part is described by (until DESCRIPTION END):
%s
DESCRIPTION END
"""
            % self.ai_config["desc"]
        )

        if "properties" in self.config:
            properties = "\n".join(
                [
                    "  %s: %s" % (k, v)
                    for k, v in self.config["properties"].items()
                ]
            )
            prompt += (
                """

The part is further described by the following properties:
%s
"""
                % properties
            )

        image_filenames = self.ai_config.get("images", [])
        if len(image_filenames) > 0:
            prompt += """
            
The part is further described by the images:
"""
            for image_filename in image_filenames:
                prompt += "INSERT_IMAGE_HERE(%s)\n" % image_filename

        prompt += """

Ensure all dimensions, distances and angles specified int the input data
are reflected in the output CSG instructions.
Use milimeters for dimensions and degrees for angles.
"""

        config = self.ai_config
        config = copy.copy(config)
        # if config["temperature"] < 0.8:
        #     config["temperature"] += 0.4
        # if config["top_p"] < 0.4:
        #     config["top_p"] += 0.2
        options = self.generate(
            "Geometric",
            self.project.name,
            self.name,
            prompt,
            config,
            self.num_geometric_modeling,
        )
        return options

    def _generate_script(self, csg_instructions):
        """This method generates a script given specific CSG description."""

        prompt = """You are an AI assistant in an engineering department.
You are helping engineers by writing scripts that produce CAD geometry data
for parts, mechanisms, buildings or anything else.
The scripts you create are fully functional and can be used right away, as is, in automated workflows.
Assume that the scripts you produce are used automatically to render 3D models and to validate them.
This time you are asked to generate a %s to define a 3D model of a part defined by
the following CSG instructions (until CSG END):
%s
CSG END
Ensure that all primitives are placed in the correct coordinates and that all dimensions are correct.
""" % (
            self.script_type,
            csg_instructions,
        )

        image_filenames = self.ai_config.get("images", [])
        if len(image_filenames) > 0:
            prompt += """
            
The part is further described by the images:
"""
            for image_filename in image_filenames:
                prompt += "INSERT_IMAGE_HERE(%s)\n" % image_filename

        prompt += """%s

IMPORTANT: Output the %s itself and do not add any text or comments before or after it.
""" % (
            self.prompt_suffix,
            self.script_type,
        )

        config = self.ai_config
        config = copy.copy(config)
        # if config["temperature"] < 0.8:
        #     config["temperature"] += 0.4
        # if config["top_p"] < 0.4:
        #     config["top_p"] += 0.2
        scripts = self.generate(
            "Script",
            self.project.name,
            self.name,
            prompt,
            config,
        )

        # Sanitize the output to remove the decorations
        scripts = list(map(lambda s: self._sanitize_script(s), scripts))

        return scripts

    def _change_script(
        self, csg_instructions, script, rendered_image, change=None
    ):
        """This method changes the script given the original request and the produced script."""

        config = copy.copy(self.ai_config)

        prompt = """You are an AI assistant in an engineering department.
You are asked to create a %s matching the given description%s.

The given description follows (until DESCRIPTION END): 
%s
DESCRIPTION END
""" % (
            self.script_type,
            " and images" if len(config["images"]) > 0 else "",
            config["desc"],
        )

        image_filenames = config["images"]
        if len(image_filenames) > 0:
            prompt += """

The given images are:
"""
            for image_filename in image_filenames:
                prompt += "INSERT_IMAGE_HERE(%s)\n" % image_filename

        if csg_instructions is not None:
            prompt += (
                """

You considered the following constructive solid geometry model (until CSG END):
%s
CSG END
"""
                % csg_instructions
            )

        prompt += (
            """
You produced the following script (until SCRIPT END):
%s
SCRIPT END
"""
            % script
        )

        prompt += """
When rendered, this script produces the following image:
"""
        prompt += "INSERT_IMAGE_HERE(%s)\n" % rendered_image

        if change is not None:
            prompt += f"\n\n{change}\n"
        else:
            prompt += """

Please, analyze whether the produced script and image match the original request
(where the original image and description take precedence
over the constructive solid geometry instructions).
Analyze both the shape and the dimensions.
Pay special attention to the coordinates used to place the initial geometric primitives.
Make sure every single dimension provided in the request are reflected in the produced script.
It's okay for some coordinates and offsets to be negative.
"""

        prompt += """
If they do precisely match the request, repeat the same script.
Otherwise, produce a corrected script following the instructions:
Do not generate exactly the same script
(make the changes necessary to address identified issues).
%s
""" % (
            self.prompt_suffix,
        )

        if config["temperature"] < 0.8:
            config["temperature"] += 0.8
        if config["top_p"] < 0.4:
            config["top_p"] += 0.4

        scripts = self.generate(
            "Change",
            self.project.name,
            self.name,
            prompt,
            config,
            self.num_script_correction,  # TODO(clairbee): add a separate user config param and loop around this until the needed number is produced
        )

        # Sanitize the output to remove the decorations
        scripts = list(map(lambda s: self._sanitize_script(s), scripts))

        return scripts

    def _sanitize_script(self, script: str):
        """Cleans up the GenAI output to keep the code only."""

        # Extract the first block between ``` if any
        loc_1 = script.find("```")
        loc_2 = script[loc_1 + 1 :].find("```") + loc_1
        if loc_2 > 0:
            # Note: the first ``` (and whatever follows on the same line) is still included
            script = script[loc_1:loc_2]

        # Remove ```` if anything is left
        script = "\n".join(
            list(
                filter(
                    lambda l: (
                        False
                        if l.startswith("```") or l.startswith(" ```")
                        else True
                    ),
                    script.split("\n"),
                )
            )
        )

        if hasattr(self, "lang") and self.lang == self.LANG_PYTHON:
            # Strip straight to the import statements (for AIs that don't ```)
            if script.startswith("from "):
                loc_from = 0
            else:
                loc_from = script.find("\nfrom ")
            if script.startswith("import "):
                loc_import = 0
            else:
                loc_import = script.find("\nimport ")

            if loc_from != -1 and loc_import != -1:
                loc = min(loc_import, loc_from)
            elif loc_from != -1:
                loc = loc_from
            else:
                loc = loc_import

            if loc != -1:
                script = script[loc:]

        return script

    def _validate_and_fix(self, script, candidate_id, depth=0):
        """
        Validate the image by rendering it,
        attempt to correct the script if rendering doesn't work.
        """
        next_depth = depth + 1

        # Render this script into an image.
        # We can't just stop at validating geometry,
        # as we need to feed the picture into the following AI logic.
        image_filename, error_text = self._render_image(script, candidate_id)
        if image_filename is not None:
            return image_filename, script
        # Failed to render the image.

        if next_depth <= self.num_script_correction and error_text:
            # Ask AI to make incremental fixes based on the errors.
            correction_candidate_id = 0
            for _ in range(self.num_script_correction):
                corrected_scripts = self._correct_script(script, error_text)
                corrected_scripts = list(
                    map(lambda s: self._sanitize_script(s), corrected_scripts)
                )
                for corrected_script in corrected_scripts:
                    pc_logging.debug(
                        "Corrected the script candidate %d, correction candidate %d at depth %d: %s"
                        % (
                            candidate_id,
                            correction_candidate_id,
                            depth,
                            corrected_script,
                        )
                    )

                    image_filename, corrected_script = self._validate_and_fix(
                        corrected_script,
                        candidate_id,
                        next_depth,
                    )
                    if image_filename is not None:
                        return image_filename, corrected_script

                    correction_candidate_id += 1

        return None, script

    def _correct_script(self, script, error_text):
        # TODO(clairbee): prove that the use of CSG instructions product
        #                 in this prompt is benefitial
        prompt = """You are an AI assistant to a mechanical engineer.
You are given an automatically generated %s which has flaws that need to be
corrected.

The generated script that contains errors is (until SCRIPT_END):
```
%s
```
SCRIPT_END

When the given script is executed, the following error messages are
produced (until ERRORS_END):
%s
ERRORS_END

Please, generate a corrected script so that it does not produce the given errors.
Limit the changes to the script to the minimum necessary to fix the errors.
Very important not to produce exactly the same script: at least something has to change.
""" % (
            self.script_type,
            script,
            error_text,
        )

        pc_logging.debug("Correction prompt: %s" % prompt)

        options = self.generate(
            "ScriptIncr",
            self.project.name,
            self.name,
            prompt,
            self.ai_config,
            self.num_script_correction,
        )
        return options

    def _render_image(self, script, candidate_id):
        """This method ensures the validity of a script candidate by attempting to render it."""
        error_text = ""
        exception_text = ""

        source_path = tempfile.mktemp(suffix=self.extension)
        with open(source_path, "w") as f:
            f.write(script)
            f.close()
        pc_logging.debug("Generated script: %s" % source_path)

        output_path = tempfile.mktemp(suffix=".png")
        pc_logging.debug("Attempting to render: %s" % output_path)

        # Create an ephemeral part for the given script candidate
        part = self._create_part(
            {
                "name": self.name + "_candidate_" + str(candidate_id),
                # "path": source_path, # TODO(clairbee): doesn't this work? why?
                "type": self.part_type,
                "mute": True,  # Suppress errors
            }
        )
        # Since this is an ephemeral part, certain things need to be tweaked
        part.instantiate = self.instantiate_orig  # Remove the AI wrapper
        part.path = source_path  # Set the path to the temporary script file
        pc_logging.debug("Part created: %.2f KB" % (total_size(part) / 1024.0))

        def render(part):
            nonlocal exception_text
            try:
                coro = part.get_shape()
                with pc_logging.Action(
                    "Instantiate", part.project_name, part.name
                ):
                    shape = asyncio.run(coro)
                if not shape is None:
                    try:
                        # Best effort to provide an interactive experience
                        part.show()
                    except Exception as e:
                        pass
                    part.render_png(self.ctx, None, output_path)
            except Exception as e:
                part.error("Failed to render the image: %s" % e)
                exception_text += f"Exception:\n{str(e)}\n"

        # TODO(clairbee): make it async up until this point, drop threads
        # Given we don't know whether the current thread is already
        # running an event loop or not, we need to create a new thread
        t = threading.Thread(target=render, args=[part])
        t.start()
        t.join()

        errors = part.errors  # Save the errors
        errors = list(
            filter(
                lambda e: "Exception while deserializing" not in e,
                errors,
            )
        )
        if len(errors) > 0:
            pc_logging.debug(
                "There were errors while attemtping to render the image"
            )
            pc_logging.debug("%s" % errors)
            for error in errors:
                error_text += f"{error}\n"

        error_text = error_text + exception_text
        os.unlink(source_path)
        if not os.path.exists(output_path) or os.path.getsize(output_path) == 0:
            pc_logging.info(
                "Script candidate %d: failed to render the image" % candidate_id
            )
            return None, error_text
        pc_logging.info(
            "Script candidate %d: successfully rendered the image"
            % candidate_id
        )
        return output_path, error_text

    def select_best_image(self, script_candidates, change=None):
        """Iterate over script_candidates and compare images.
        Each script candidate is [<image_filename>, <script>]."""

        if len(script_candidates) == 0:
            return None

        image_filenames = list(map(lambda c: c[0], script_candidates))

        prompt = (
            """
You are an AI assistant to a mechanical engineer.
The mechanical engineer was given a task to create a 3D model of a part.
The engineer has produced several models and rendered an image per model.

The part is described as (until DESCRIPTION END):
%s
DESCRIPTION END
"""
            % self.ai_config["desc"]
        )

        if "images" in self.ai_config and len(self.ai_config["images"]) > 0:
            prompt += """
The part is further described by the images:
"""
            for image_filename in self.ai_config["images"]:
                prompt += "INSERT_IMAGE_HERE(%s)\n" % image_filename

        if change is not None:
            prompt += (
                """

Subsequently, the following changes were requested (until "CHANGE END"):
%s
CHANGE END
"""
                % change
            )

        prompt += """

From the following images,
select the rendered image that matches the requirements the best:
"""
        for image_filename in image_filenames:
            prompt += "INSERT_IMAGE_HERE(%s)\n" % image_filename

        prompt += """

Respond with the numeric index (starting with 1) of the rendered image
that fits the requirements the best.
No other text is acceptable.
Just the number.
"""

        # Ask AI to compare the images
        pc_logging.info(
            "Attempting to select the best script by comparing images"
        )
        config = copy.copy(self.ai_config)
        if config["temperature"] > 0.1:
            config["temperature"] = 0.05
        if config["top_p"] > 0.1:
            config["top_p"] = 0.05
        responses = self.generate(
            "Compare",
            self.project.name,
            self.name,
            prompt,
            config,
            1,
        )

        pc_logging.debug("Image comparison responses: %s" % responses)

        # Parse the response to get the number out
        for response in responses:
            words = response.split(" ")
            words = list(filter(lambda w: w.strip().isnumeric(), words))
            if len(words) == 1:
                idx = int(words[0])
                if idx > 0 and idx <= len(script_candidates):
                    pc_logging.info("Selected the best fit image: %d" % idx)
                    return script_candidates[idx - 1][1]

        return script_candidates[0][1]



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-03-23
#
# Licensed under Apache License, Version 2.0.
#

from .part_factory_build123d import PartFactoryBuild123d
from .part_factory_feature_ai import PartFactoryFeatureAi
from . import logging as pc_logging


class PartFactoryAiBuild123d(PartFactoryBuild123d, PartFactoryFeatureAi):
    def __init__(self, ctx, source_project, target_project, config):
        # Override the path determined by the parent class to enable "enrich"
        config["path"] = config["name"] + ".py"
        self.lang = self.LANG_PYTHON

        mode = "builder"
        if "mode" in config and config["mode"] == "algebra":
            mode = "algebra"

        with pc_logging.Action(
            "InitAiB3d", target_project.name, config["name"]
        ):
            PartFactoryFeatureAi.__init__(
                self,
                config,
                "build123d",
                "build123d script (in %s)" % mode,
                """Import all the required modules
(including 'math' and 'build123d' itself).
Do not export anything.
Use show_object() to display the part.
""",
            )
            PartFactoryBuild123d.__init__(
                self,
                ctx,
                source_project,
                target_project,
                config,
                can_create=True,
            )

            self.on_init_ai()



#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-02-19
#
# Licensed under Apache License, Version 2.0.
#

from . import logging as pc_logging


class Factory:
    def __init__(self) -> None:
        pass


all = {
    "assembly": {},
    "part": {},
    "file": {},
    "provider": {},
}


def register(kind: str, t: str, factory_class: Factory.__class__):
    all[kind][t] = factory_class


def instantiate(kind: str, t: str, ctx, source_project, target_project, config):
    if t in all[kind]:
        # The return value is not always used
        return all[kind][t](ctx, source_project, target_project, config)
    else:
        pc_logging.error("Invalid %s type encountered: %s" % (kind, config))



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-01-06
#
# Licensed under Apache License, Version 2.0.
#

import asyncio
import os
import shutil
import subprocess
import tempfile

import build123d as b3d

from .part_factory_file import PartFactoryFile
from . import logging as pc_logging


class PartFactoryScad(PartFactoryFile):
    def __init__(
        self, ctx, source_project, target_project, config, can_create=False
    ):
        with pc_logging.Action(
            "InitOpenSCAD", target_project.name, config["name"]
        ):
            super().__init__(
                ctx,
                source_project,
                target_project,
                config,
                extension=".scad",
                can_create=can_create,
            )
            # Complement the config object here if necessary
            self._create(config)

            self.project_dir = source_project.config_dir

    async def instantiate(self, part):
        await super().instantiate(part)

        with pc_logging.Action("OpenSCAD", part.project_name, part.name):
            if not os.path.exists(part.path) or os.path.getsize(part.path) == 0:
                pc_logging.error(
                    "OpenSCAD script is empty or does not exist: %s" % part.path
                )
                return None

            scad_path = shutil.which("openscad")
            if scad_path is None:
                raise Exception(
                    "OpenSCAD executable is not found. Please, install OpenSCAD first."
                )

            stl_path = tempfile.mktemp(".stl")
            p = await asyncio.create_subprocess_exec(
                *[
                    scad_path,
                    "--export-format",
                    "binstl",
                    "-o",
                    stl_path,
                    part.path,
                ],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            _, errors = await p.communicate()
            if len(errors) > 0:
                error_lines = errors.decode().split("\n")
                for error_line in error_lines:
                    pc_logging.debug("%s: %s" % (part.name, error_line))

            if not os.path.exists(stl_path) or os.path.getsize(stl_path) == 0:
                part.error(
                    "OpenSCAD failed to generate the STL file. Please, check the script."
                )
                return None

            try:
                shape = b3d.Mesher().read(stl_path)[0].wrapped
            except:
                try:
                    # First, make sure it's not the known problem in Mesher
                    shape = b3d.import_stl(stl_path).wrapped
                except Exception as e:
                    part.error("%s: %s" % (part.name, e))
                    return None
            os.unlink(stl_path)

            self.ctx.stats_parts_instantiated += 1

            return shape



# Original code taken from https://gist.github.com/SDI8/3137ee70649e4901913c7c8e6b534ec8

"""
MIT License
Copyright (c) 2022 Simon Dibbern
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

# Also influenced by https://github.com/gumyr/build123d/blob/dev/src/build123d/persistence.py

"""
build123d pickle support

name: persistence.py
by:   Jojain & bernhard-42
date: September 8th, 2023

desc:
    This python module enables build123d objects to be pickled.

license:

    Copyright 2023 Jojain & bernhard-42

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""

import copyreg
from io import BytesIO
from typing import Any

import cadquery as cq
import OCP
from OCP.TopAbs import TopAbs_Orientation, TopAbs_ShapeEnum
from OCP.BinTools import BinTools
from OCP.gp import gp_Quaternion, gp_Trsf, gp_Vec
from OCP.TopLoc import TopLoc_Location
from OCP.TopoDS import (
    TopoDS,
    TopoDS_Compound,
    TopoDS_CompSolid,
    TopoDS_Edge,
    TopoDS_Face,
    TopoDS_Shape,
    TopoDS_Shell,
    TopoDS_Solid,
    TopoDS_Vertex,
    TopoDS_Wire,
)

import OCP.TopAbs as ta

downcast_LUT = {
    ta.TopAbs_VERTEX: TopoDS.Vertex_s,
    ta.TopAbs_EDGE: TopoDS.Edge_s,
    ta.TopAbs_WIRE: TopoDS.Wire_s,
    ta.TopAbs_FACE: TopoDS.Face_s,
    ta.TopAbs_SHELL: TopoDS.Shell_s,
    ta.TopAbs_SOLID: TopoDS.Solid_s,
    ta.TopAbs_COMPOUND: TopoDS.Compound_s,
    ta.TopAbs_COMPSOLID: TopoDS.CompSolid_s,
}


def shapetype(obj: TopoDS_Shape) -> TopAbs_ShapeEnum:
    """Return TopoDS_Shape's TopAbs_ShapeEnum"""
    if obj.IsNull():
        raise ValueError("Null TopoDS_Shape object")

    return obj.ShapeType()


def downcast(obj: TopoDS_Shape) -> TopoDS_Shape:
    """Downcasts a TopoDS object to suitable specialized type

    Args:
      obj: TopoDS_Shape:

    Returns:

    """

    f_downcast: Any = downcast_LUT[shapetype(obj)]
    return_value = f_downcast(obj)

    return return_value


def _inflate_shape(data: bytes):
    with BytesIO(data) as bio:
        return cq.Shape.importBrep(bio)


def _reduce_shape(shape: cq.Shape):
    with BytesIO() as stream:
        shape.exportBrep(stream)
        return _inflate_shape, (stream.getvalue(),)


def _inflate_topods(data: bytes):
    with BytesIO(data) as bio:
        shape = TopoDS_Shape()
        builder = OCP.BRep.BRep_Builder()
        OCP.BRepTools.BRepTools.Read_s(shape, bio, builder)
        return downcast(shape)


def _reduce_topods(shape):
    with BytesIO() as bio:
        OCP.BRepTools.BRepTools.Write_s(shape, bio)
        return _inflate_topods, (bio.getvalue(),)


def _inflate_transform(*values: float):
    trsf = OCP.gp.gp_Trsf()
    trsf.SetValues(*values)
    return trsf


def _reduce_transform(transform: OCP.gp.gp_Trsf):
    return _inflate_transform, tuple(
        transform.Value(i, j) for i in range(1, 4) for j in range(1, 5)
    )


def _inflate_Gtransform(*values):
    gtrsf = OCP.gp.gp_GTrsf(values[0], values[1])
    return gtrsf


def _reduce_Gtransform(transform: OCP.gp.gp_GTrsf):
    return _inflate_Gtransform, tuple(
        [
            transform.VectorialPart(),
            transform.TranslationPart(),
        ]
    )


def _inflate_mat(*values):
    trsf = OCP.gp.gp_Mat(values[0], values[1], values[2])
    return trsf


def _reduce_mat(mat: OCP.gp.gp_Mat):
    return _inflate_mat, (
        mat.Column(1),
        mat.Column(2),
        mat.Column(3),
    )


def _inflate_ax1(*values: float):
    ax1 = OCP.gp.gp_Ax1(values[0], values[1])
    return ax1


def _reduce_ax1(ax1: OCP.gp.gp_Ax1):
    return _inflate_ax1, (
        ax1.Location(),
        ax1.Direction(),
    )


def _inflate_ax3(*values: float):
    ax3 = OCP.gp.gp_Ax3(values[0], values[1], values[2])
    ax3.SetYDirection(values[3])
    return ax3


def _reduce_ax3(ax3: OCP.gp.gp_Ax3):
    return _inflate_ax3, (
        ax3.Location(),
        ax3.Direction(),
        ax3.XDirection(),
        ax3.YDirection(),
    )


def _inflate_pln(*values: float):
    ax3 = OCP.gp.gp_Ax3(values[0], values[1], values[2])
    ax3.SetYDirection(values[3])
    pln = OCP.gp.gp_Pln(ax3)
    return pln


def _reduce_pln(pln: OCP.gp.gp_Pln):
    ax3 = pln.Position()
    return _inflate_pln, (
        ax3.Location(),
        ax3.Direction(),
        ax3.XDirection(),
        ax3.YDirection(),
    )


def _inflate_pnt(*values: float):
    pnt = OCP.gp.gp_Pnt(values[0], values[1], values[2])
    return pnt


def _reduce_pnt(pnt: OCP.gp.gp_Pnt):
    return _inflate_pnt, (pnt.X(), pnt.Y(), pnt.Z())


def _inflate_vec(*values: float):
    pnt = OCP.gp.gp_Vec(values[0], values[1], values[2])
    return pnt


def _reduce_vec(pnt: OCP.gp.gp_Vec):
    return _inflate_vec, (pnt.X(), pnt.Y(), pnt.Z())


def _inflate_dir(*values: float):
    dir = OCP.gp.gp_Dir(values[0], values[1], values[2])
    return dir


def _reduce_dir(dir: OCP.gp.gp_Dir):
    return _inflate_dir, (dir.X(), dir.Y(), dir.Z())


def _inflate_xyz(*values: float):
    dir = OCP.gp.gp_XYZ(values[0], values[1], values[2])
    return dir


def _reduce_xyz(dir: OCP.gp.gp_XYZ):
    return _inflate_xyz, (dir.X(), dir.Y(), dir.Z())


def register():
    """
    Registers pickle support functions for common CadQuery and OCCT objects.
    """

    for cls in (
        cq.Edge,
        cq.Compound,
        cq.Shell,
        cq.Face,
        cq.Solid,
        cq.Vertex,
        cq.Wire,
    ):
        copyreg.pickle(cls, _reduce_shape)

    copyreg.pickle(OCP.gp.gp_Pnt, _reduce_pnt)
    copyreg.pickle(OCP.gp.gp_Vec, _reduce_vec)
    copyreg.pickle(OCP.gp.gp_Dir, _reduce_dir)
    copyreg.pickle(OCP.gp.gp_Mat, _reduce_mat)
    copyreg.pickle(OCP.gp.gp_XYZ, _reduce_xyz)

    copyreg.pickle(OCP.gp.gp_Ax1, _reduce_ax1)
    copyreg.pickle(OCP.gp.gp_Ax3, _reduce_ax3)
    copyreg.pickle(OCP.gp.gp_Pln, _reduce_pln)

    copyreg.pickle(cq.Vector, lambda vec: (cq.Vector, vec.toTuple()))
    copyreg.pickle(OCP.gp.gp_Trsf, _reduce_transform)
    copyreg.pickle(OCP.gp.gp_GTrsf, _reduce_Gtransform)
    copyreg.pickle(
        cq.Location, lambda loc: (cq.Location, (loc.wrapped.Transformation(),))
    )
    copyreg.pickle(
        TopLoc_Location,
        lambda loc: (TopLoc_Location, (loc.Transformation(),)),
    )

    for cls in (
        TopoDS_Shape,
        TopoDS_Compound,
        TopoDS_CompSolid,
        TopoDS_Solid,
        TopoDS_Shell,
        TopoDS_Face,
        TopoDS_Wire,
        TopoDS_Edge,
        TopoDS_Vertex,
    ):
        copyreg.pickle(cls, _reduce_topods)



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-09-07
#
# Licensed under Apache License, Version 2.0.
#

import typing

from .provider_request_caps import ProviderRequestCaps
from .provider_request_order import ProviderRequestOrder
from .provider_request_quote import ProviderRequestQuote
from .provider_data_cart import *
from . import logging as pc_logging


class Provider:
    name: str
    config: dict[str, typing.Any] = None
    path: typing.Optional[str] = None
    url: typing.Optional[str] = None
    errors: list[str]
    caps: dict[str, typing.Any] = None

    def __init__(self, name: str, config: dict[str, typing.Any] = {}):
        super().__init__()
        self.name = name
        self.config = config

        self.url = config.get("url", None)

    def error(self, msg: str):
        mute = self.config.get("mute", False)
        if mute:
            self.errors.append(msg)
        else:
            pc_logging.error(msg)

    def is_qos_available(self, qos: str) -> bool:
        # TODO(clairbee): use the config to determine if the QoS is available
        return True

    async def get_caps(self) -> dict[str, typing.Any]:
        if not self.caps:
            self.caps = await self.query_caps(ProviderRequestCaps())
        return self.caps

    async def is_part_available(self, cart_item: ProviderCartItem):
        raise NotImplementedError()

    async def load(self, cart_item: ProviderCartItem):
        raise NotImplementedError()

    async def query_caps(self, request: ProviderRequestCaps):
        raise NotImplementedError()

    async def query_quote(self, request: ProviderRequestQuote):
        raise NotImplementedError()

    async def query_order(self, request: ProviderRequestOrder):
        raise NotImplementedError()



#
# OpenVMP, 2024
#
# Author: Roman Kuzmenko
# Created: 2024-04-20
#
# Licensed under Apache License, Version 2.0.
#

import build123d as b3d

from OCP.ShapeExtend import ShapeExtend_WireData
from OCP.ShapeFix import (
    ShapeFix_Shape,
)
from OCP.BRepAlgoAPI import BRepAlgoAPI_Fuse
from OCP.BOPAlgo import BOPAlgo_Operation
from OCP.TopTools import TopTools_ListOfShape

from OCP.BRepBuilderAPI import BRepBuilderAPI_MakeFace
from OCP.TopoDS import TopoDS

from . import logging as pc_logging
from .sketch_factory_file import SketchFactoryFile


class SketchFactorySvg(SketchFactoryFile):
    flip_y = True
    ignore_visibility = False
    use_wires = False
    use_faces = False

    def __init__(self, ctx, source_project, target_project, config):
        with pc_logging.Action("InitSVG", target_project.name, config["name"]):
            super().__init__(
                ctx,
                source_project,
                target_project,
                config,
                extension=".svg",
            )

            if "flip-y" in config:
                self.flip_y = bool(config["flip-y"])

            if "ignore-visibility" in config:
                self.ignore_visibility = bool(config["ignore-visibility"])

            if "use-wires" in config:
                self.use_wires = bool(config["use-wires"])

            if "use-faces" in config:
                self.use_faces = bool(config["use-faces"])

            self._create(config)

    async def instantiate(self, sketch):
        await super().instantiate(sketch)

        with pc_logging.Action("SVG", sketch.project_name, sketch.name):
            try:
                shape_list = b3d.import_svg(
                    self.path,
                    flip_y=self.flip_y,
                    ignore_visibility=self.ignore_visibility,
                )

                faces = []
                if self.use_wires or not self.use_faces:
                    wires = shape_list.wires()

                    wire_merger = ShapeExtend_WireData()
                    for wire in wires:
                        wire_merger.Add(wire.wrapped)
                    wire = wire_merger.Wire()

                    wire_fixer = ShapeFix_Shape(wire)
                    wire_fixer.Perform()
                    fixed_wire_shape = wire_fixer.Shape()
                    wire = TopoDS.Wire_s(fixed_wire_shape)

                    face_builder = BRepBuilderAPI_MakeFace(wire, True)
                    face_builder.Build()
                    if not face_builder.IsDone():
                        raise ValueError(
                            f"Cannot build face(s): {face_builder.Error()}"
                        )

                    face = face_builder.Face()

                    face_fixer = ShapeFix_Shape(face)
                    face_fixer.Perform()
                    fixed_face_shape = face_fixer.Shape()
                    face = TopoDS.Face_s(fixed_face_shape)

                    faces.append(face)

                if self.use_faces or not self.use_wires:
                    faces.extend(shape_list.faces())

                if len(faces) == 1:
                    shape = faces[0]
                else:
                    # TODO(clairbee): verify this branch
                    shapes = TopTools_ListOfShape()
                    for face in faces:
                        shapes.Append(face)
                    face_fuser = BRepAlgoAPI_Fuse()
                    face_fuser.SetArguments(shapes)
                    face_fuser.SetOperation(BOPAlgo_Operation.BOPAlgo_FUSE)
                    # face_fuser.SetRunParallel(True)
                    face_fuser.Build()
                    if face_fuser.IsDone():
                        shape = face_fuser.Shape()
                    else:
                        shape = None
            except Exception as e:
                pc_logging.exception(
                    "Failed to import the SVG file: %s: %s" % (self.path, e)
                )
                shape = None

            self.ctx.stats_sketches_instantiated += 1

            return shape
